from ....module.model import EncoderUniversalRegistry, DecoderUniversalRegistry, ModelUniversalRegistry

def _parse_hp_space(spaces):
    if spaces is None:
        return None
    for space in spaces:
        if "cutFunc" in space and isinstance(space["cutFunc"], str):
            space["cutFunc"] = eval(space["cutFunc"])
    return spaces

def _parse_model_hp(model):
    assert isinstance(model, dict)
    if "encoder" in model and "decoder" in model:
        if "prediction_head" in model:
            return {
                "encoder": _parse_hp_space(model["encoder"].pop("hp_space", None)),
                "decoder": _parse_hp_space(model["decoder"].pop("hp_space", None)),
                "prediction_head": _parse_hp_space(model["prediction_head"].pop("hp_space", None))
            }
        else:
            return {
                "encoder": _parse_hp_space(model["encoder"].pop("hp_space", None)),
                "decoder": _parse_hp_space(model["decoder"].pop("hp_space", None)),
            }
    elif "encoder" in model:
        return {
            "encoder": _parse_hp_space(model["encoder"].pop("hp_space", None)),
            "decoder": None,
        }
    else:
        return _parse_hp_space(model.pop("hp_space", None))

def _initialize_single_model(model):
    encoder, decoder = None, None
    if "encoder" in model:
        # initialize encoder
        name = model["encoder"].pop("name")
        encoder = EncoderUniversalRegistry.get_encoder(name)(**model["encoder"])
    if "decoder" in model:
        # initialize decoder
        name = model["decoder"].pop("name")
        decoder = DecoderUniversalRegistry.get_decoder(name)(**model["decoder"])
        return (encoder, decoder)

    if "name" in model:
        # whole model
        name = model.pop("name")
        encoder = ModelUniversalRegistry.get_model(name)(**model)
    return encoder



import torch
import torch.nn.functional as F
from torch.nn import Linear, ReLU, Sequential, LeakyReLU, Tanh, ELU
from torch_geometric.nn import GINConv, global_add_pool
from torch.nn import BatchNorm1d
from . import register_model
from .base import BaseAutoModel, activate_func
from copy import deepcopy
from ....utils import get_logger

LOGGER = get_logger("GINModel")


def set_default(args, d):
    for k, v in d.items():
        if k not in args:
            args[k] = v
    return args


class GIN(torch.nn.Module):
    def __init__(self, args):
        super(GIN, self).__init__()
        self.args = args
        self.num_layer = int(self.args["num_layers"])
        assert self.num_layer > 2, "Number of layers in GIN should not less than 3"

        missing_keys = list(
            set(
                [
                    "features_num",
                    "num_class",
                    "num_graph_features",
                    "num_layers",
                    "hidden",
                    "dropout",
                    "act",
                    "mlp_layers",
                    "eps",
                ]
            )
            - set(self.args.keys())
        )
        if len(missing_keys) > 0:
            raise Exception("Missing keys: %s." % ",".join(missing_keys))
        if not self.num_layer == len(self.args["hidden"]) + 1:
            LOGGER.warn("Warning: layer size does not match the length of hidden units")
        self.num_graph_features = self.args["num_graph_features"]

        if self.args["act"] == "leaky_relu":
            act = LeakyReLU()
        elif self.args["act"] == "relu":
            act = ReLU()
        elif self.args["act"] == "elu":
            act = ELU()
        elif self.args["act"] == "tanh":
            act = Tanh()
        else:
            act = ReLU()

        train_eps = True if self.args["eps"] == "True" else False

        self.convs = torch.nn.ModuleList()
        self.bns = torch.nn.ModuleList()

        nn = [Linear(self.args["features_num"], self.args["hidden"][0])]
        for _ in range(self.args["mlp_layers"] - 1):
            nn.append(act)
            nn.append(Linear(self.args["hidden"][0], self.args["hidden"][0]))
        # nn.append(BatchNorm1d(self.args['hidden'][0]))
        self.convs.append(GINConv(Sequential(*nn), train_eps=train_eps))
        self.bns.append(BatchNorm1d(self.args["hidden"][0]))

        for i in range(self.num_layer - 3):
            nn = [Linear(self.args["hidden"][i], self.args["hidden"][i + 1])]
            for _ in range(self.args["mlp_layers"] - 1):
                nn.append(act)
                nn.append(
                    Linear(self.args["hidden"][i + 1], self.args["hidden"][i + 1])
                )
            # nn.append(BatchNorm1d(self.args['hidden'][i+1]))
            self.convs.append(GINConv(Sequential(*nn), train_eps=train_eps))
            self.bns.append(BatchNorm1d(self.args["hidden"][i + 1]))

        self.fc1 = Linear(
            self.args["hidden"][self.num_layer - 3] + self.num_graph_features,
            self.args["hidden"][self.num_layer - 2],
        )
        self.fc2 = Linear(
            self.args["hidden"][self.num_layer - 2], self.args["num_class"]
        )

    def forward(self, data):
        x, edge_index, batch = data.x, data.edge_index, data.batch

        if self.num_graph_features > 0:
            graph_feature = data.gf

        for i in range(self.num_layer - 2):
            x = self.convs[i](x, edge_index)
            x = activate_func(x, self.args["act"])
            x = self.bns[i](x)

        x = global_add_pool(x, batch)
        if self.num_graph_features > 0:
            x = torch.cat([x, graph_feature], dim=-1)
        x = self.fc1(x)
        x = activate_func(x, self.args["act"])
        x = F.dropout(x, p=self.args["dropout"], training=self.training)

        x = self.fc2(x)

        return F.log_softmax(x, dim=1)


@register_model("gin-model")
class AutoGIN(BaseAutoModel):
    r"""
    AutoGIN. The model used in this automodel is GIN, i.e., the graph isomorphism network from the `"How Powerful are
    Graph Neural Networks?" <https://arxiv.org/abs/1810.00826>`_ paper. The layer is

    .. math::
        \mathbf{x}^{\prime}_i = h_{\mathbf{\Theta}} \left( (1 + \epsilon) \cdot
        \mathbf{x}_i + \sum_{j \in \mathcal{N}(i)} \mathbf{x}_j \right)

    or

    .. math::
        \mathbf{X}^{\prime} = h_{\mathbf{\Theta}} \left( \left( \mathbf{A} +
        (1 + \epsilon) \cdot \mathbf{I} \right) \cdot \mathbf{X} \right),

    here :math:`h_{\mathbf{\Theta}}` denotes a neural network, *.i.e.* an MLP.

    Parameters
    ----------
    num_features: `int`.
        The dimension of features.

    num_classes: `int`.
        The number of classes.

    device: `torch.device` or `str`
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.
    """

    def __init__(
        self,
        num_features=None,
        num_classes=None,
        device=None,
        init=False,
        num_graph_features=0,
        **args
    ):

        super().__init__(num_features, num_classes, device, num_graph_features=num_graph_features, **args)
        self.num_graph_features = num_graph_features

        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "4,5,6",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 5,
                "minValue": [8, 8, 8, 8, 8],
                "maxValue": [64, 64, 64, 64, 64],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
            {
                "parameterName": "eps",
                "type": "CATEGORICAL",
                "feasiblePoints": ["True", "False"],
            },
            {
                "parameterName": "mlp_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
        ]

        self.hyper_parameters = {
            "num_layers": 3,
            "hidden": [64, 32],
            "dropout": 0.5,
            "act": "relu",
            "eps": "True",
            "mlp_layers": 2,
        }

    def from_hyper_parameter(self, hp, **kwargs):
        return super().from_hyper_parameter(hp, num_graph_features=self.num_graph_features, **kwargs)

    def _initialize(self):
        # """Initialize model."""
        self._model = GIN({
            "features_num": self.input_dimension,
            "num_class": self.output_dimension,
            "num_graph_features": self.num_graph_features,
            **self.hyper_parameters
        }).to(self.device)



from .base_decoder import BaseDecoderMaintainer
from .decoder_registry import DecoderUniversalRegistry
from autogl.backend import DependentBackend

if DependentBackend.is_pyg():
    from ._pyg import (
        LogSoftmaxDecoderMaintainer,
        SumPoolMLPDecoderMaintainer,
        DiffPoolDecoderMaintainer,
        DotProductLinkPredictionDecoderMaintainer
    )
else:
    from ._dgl import (
        LogSoftmaxDecoderMaintainer,
        JKSumPoolDecoderMaintainer,
        TopKDecoderMaintainer,
        DotProductLinkPredictionDecoderMaintainer
    )

__all__ = [
    "BaseDecoderMaintainer",
    "DecoderUniversalRegistry",
    "LogSoftmaxDecoderMaintainer",
    "DotProductLinkPredictionDecoderMaintainer"
]

if DependentBackend.is_pyg():
    __all__.extend([
        "DiffPoolDecoderMaintainer",
        "SumPoolMLPDecoderMaintainer"
    ])
else:
    __all__.extend([
        "JKSumPoolDecoderMaintainer",
        "TopKDecoderMaintainer"
    ])



import netlsd
import networkx
import torch
import typing
from autogl.data.graph import GeneralStaticGraph
from autogl.data.graph.utils import conversion
from .._feature_engineer import FeatureEngineer
from ..._data_preprocessor_registry import DataPreprocessorUniversalRegistry


@DataPreprocessorUniversalRegistry.register_data_preprocessor("NetLSD".lower())
class NetLSD(FeatureEngineer):
    r"""
    Notes
    -----
    a graph feature generation method. This is a simple wrapper of NetLSD [#]_.

    References
    ----------
    ..  [#] A. Tsitsulin, D. Mottin, P. Karras, A. Bronstein, and E. Müller, “NetLSD: Hearing the shape of a graph,”
        Proc. ACM SIGKDD Int. Conf. Knowl. Discov. Data Min., pp. 2347–2356, 2018.
    """

    def __init__(self, *args, **kwargs):
        self.__args = args
        self.__kwargs = kwargs
        super(NetLSD, self).__init__()

    def __extract(self, nx_g: networkx.Graph) -> torch.Tensor:
        return torch.tensor(netlsd.heat(nx_g, *self.__args, **self.__kwargs)).view(-1)

    def __transform_homogeneous_static_graph(
            self, homogeneous_static_graph: GeneralStaticGraph
    ) -> GeneralStaticGraph:
        if not (
                homogeneous_static_graph.nodes.is_homogeneous and
                homogeneous_static_graph.edges.is_homogeneous
        ):
            raise ValueError("Provided static graph must be homogeneous")
        dsc: torch.Tensor = self.__extract(
            conversion.HomogeneousStaticGraphToNetworkX(to_undirected=True).__call__(
                homogeneous_static_graph, to_undirected=True
            )
        )
        if 'gf' in homogeneous_static_graph.data:
            gf = homogeneous_static_graph.data['gf'].view(-1)
            homogeneous_static_graph.data['gf'] = torch.cat([gf, dsc])
        else:
            homogeneous_static_graph.data['gf'] = dsc
        return homogeneous_static_graph

    @classmethod
    def __edge_index_to_nx_graph(cls, edge_index: torch.Tensor) -> networkx.Graph:
        g: networkx.Graph = networkx.Graph()
        for u, v in edge_index.t().tolist():
            if u == v:
                continue
            else:
                g.add_edge(u, v)
        return g

    def __transform_data(self, data):
        if not (
                hasattr(data, "edge_index") and
                torch.is_tensor(data.edge_index) and
                isinstance(data.edge_index, torch.Tensor) and
                data.edge_index.dim() == data.edge_index.size(0) == 2 and
                data.edge_index.dtype == torch.long
        ):
            raise TypeError("Unsupported provided data")
        dsc: torch.Tensor = self.__extract(self.__edge_index_to_nx_graph(data.edge_index))
        if hasattr(data, 'gf') and isinstance(data.gf, torch.Tensor):
            gf = data.gf.view(-1)
            data.gf = torch.cat([gf, dsc])
        else:
            data.gf = dsc
        return data

    def _transform(
            self, data: typing.Union[GeneralStaticGraph, typing.Any]
    ) -> typing.Union[GeneralStaticGraph, typing.Any]:
        if isinstance(data, GeneralStaticGraph):
            return self.__transform_homogeneous_static_graph(data)
        else:
            return self.__transform_data(data)



import netlsd
from .base import BaseGraph
import numpy as np
import torch
from .. import register_feature


@register_feature("netlsd")
class SgNetLSD(BaseGraph):
    r"""
    Notes
    -----
    a graph feature generation method. This is a simple wrapper of NetLSD [#]_.
    References
    ----------
    .. [#] A. Tsitsulin, D. Mottin, P. Karras, A. Bronstein, and E. Müller, “NetLSD: Hearing the shape of a graph,”
     Proc. ACM SIGKDD Int. Conf. Knowl. Discov. Data Min., pp. 2347–2356, 2018.
    """

    def __init__(self, *args, **kwargs):
        super(SgNetLSD, self).__init__(data_t="nx")
        self._args = args
        self._kwargs = kwargs

    def _transform(self, data):
        dsc = torch.FloatTensor([netlsd.heat(data.G, *self._args, **self._kwargs)])
        data.gf = torch.cat([data.gf, dsc], dim=1)
        return data



import torch
import torch.nn as nn
import torch.nn.functional as F

from . import register_nas_algo
from .base import BaseNAS
from ..space import BaseSpace
from ..utils import (
    AverageMeterGroup,
    replace_layer_choice,
    replace_input_choice,
    get_module_order,
    sort_replaced_module,
)
from tqdm import tqdm
from .rl import PathSamplingLayerChoice, PathSamplingInputChoice
import numpy as np
from ....utils import get_logger

LOGGER = get_logger("random_search_NAS")


@register_nas_algo("random")
class RandomSearch(BaseNAS):
    """
    Uniformly random architecture search

    Parameters
    ----------
    device : str or torch.device
        The device of the whole process, e.g. "cuda", torch.device("cpu")
    num_epochs : int
        Number of epochs planned for training.
    disable_progeress: boolean
        Control whether show the progress bar.
    """

    def __init__(self, device="auto", num_epochs=400, disable_progress=False, hardware_metric_limit=None):
        super().__init__(device)
        self.num_epochs = num_epochs
        self.disable_progress = disable_progress
        self.hardware_metric_limit = hardware_metric_limit

    def search(self, space: BaseSpace, dset, estimator):
        self.estimator = estimator
        self.dataset = dset
        self.space = space

        self.nas_modules = []
        k2o = get_module_order(self.space)
        replace_layer_choice(self.space, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.space, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)
        selection_range = {}
        for k, v in self.nas_modules:
            selection_range[k] = len(v)
        self.selection_dict = selection_range

        # space_size=np.prod(list(selection_range.values()))

        arch_perfs = []
        cache = {}
        with tqdm(range(self.num_epochs), disable=self.disable_progress) as bar:
            for i in bar:
                selection = self.sample()
                vec = tuple(list(selection.values()))
                if vec not in cache:
                    self.arch = space.parse_model(selection, self.device)
                    metric, loss, hardware_metric = self._infer(mask="val")
                    if self.hardware_metric_limit is None or hardware_metric[0] < self.hardware_metric_limit:
                        arch_perfs.append([metric, selection])
                    cache[vec] = metric
                bar.set_postfix(acc=metric, max_acc=max(cache.values()))
        selection = arch_perfs[np.argmax([x[0] for x in arch_perfs])][1]
        arch = space.parse_model(selection, self.device)
        return arch

    def sample(self):
        # uniformly sample
        selection = {}
        for k, v in self.selection_dict.items():
            selection[k] = np.random.choice(range(v))
        return selection

    def _infer(self, mask="train"):
        metric, loss = self.estimator.infer(self.arch._model, self.dataset, mask=mask)
        return metric[0], loss, metric[1:]



import os
import typing as _typing
from autogl.utils import universal_registry
from autogl.data import Dataset


class DatasetUniversalRegistry(universal_registry.UniversalRegistryBase):
    @classmethod
    def register_dataset(cls, dataset_name: str) -> _typing.Callable[
        [_typing.Type[Dataset]], _typing.Type[Dataset]
    ]:
        def register_dataset_cls(dataset: _typing.Type[Dataset]):
            if not issubclass(dataset, Dataset):
                raise TypeError
            else:
                cls[dataset_name] = dataset
                return dataset

        return register_dataset_cls

    @classmethod
    def get_dataset(cls, dataset_name: str) -> _typing.Type[Dataset]:
        return cls[dataset_name]


def build_dataset_from_name(dataset_name: str, path: str = "~/.cache-autogl/"):
    """

    Parameters
    ----------
    dataset_name: `str`
        name of dataset
    path: `str`
        local cache directory for datasets, default to "~/.cache-autogl/"

    Returns
    -------
    instance of dataset
    """
    path = os.path.expanduser(os.path.join(path, "data", dataset_name))
    _dataset = DatasetUniversalRegistry.get_dataset(dataset_name)
    return _dataset(path)



from . import register_trainer, Evaluation
import torch
from torch.optim.lr_scheduler import StepLR, MultiStepLR, ExponentialLR, ReduceLROnPlateau
import torch.nn.functional as F
from ..model import BaseEncoderMaintainer, BaseDecoderMaintainer, BaseAutoModel
from .evaluation import Auc, EVALUATE_DICT
from .base import EarlyStopping, BaseLinkPredictionTrainer
from typing import Union, Tuple
from copy import deepcopy
from ...utils import get_logger

from ...backend import DependentBackend

LOGGER = get_logger("link prediction trainer")


def get_feval(feval):
    if isinstance(feval, str):
        return EVALUATE_DICT[feval]
    if isinstance(feval, type) and issubclass(feval, Evaluation):
        return feval
    if isinstance(feval, list):
        return [get_feval(f) for f in feval]
    raise ValueError("feval argument of type", type(feval), "is not supported!")

class _DummyLinkModel(torch.nn.Module):
    def __init__(self, encoder, decoder):
        super().__init__()
        self.automodelflag = False
        if isinstance(encoder, BaseAutoModel):
            self.automodelflag = True
            self.encoder = encoder.model
            self.decoder = None
        else:
            self.encoder = encoder.encoder
            self.decoder = None if decoder is None else decoder.decoder
    
    def encode(self, data):
        if self.automodelflag:
            return self.encoder.lp_encode(data)
        return self.encoder(data)
    
    def decode(self, features, data, pos_edges, neg_edges):
        if self.automodelflag:
            return self.encoder.lp_decode(features, pos_edges, neg_edges)
        if self.decoder is None:
            return features
        return self.decoder(features, data, pos_edges, neg_edges)

@register_trainer("LinkPredictionFull")
class LinkPredictionTrainer(BaseLinkPredictionTrainer):
    """
    The link prediction trainer.

    Used to automatically train the link prediction problem.

    Parameters
    ----------
    model:
        Models can be ``str``, ``autogl.module.model.BaseAutoModel``, 
        ``autogl.module.model.encoders.BaseEncoderMaintainer`` or a tuple of (encoder, decoder) 
        if need to specify both encoder and decoder. Encoder can be ``str`` or
        ``autogl.module.model.encoders.BaseEncoderMaintainer``, and decoder can be ``str``
        or ``autogl.module.model.decoders.BaseDecoderMaintainer``.
        If only encoder is specified, decoder will be default to "logsoftmax"

    num_features: int (Optional)
        The number of features in dataset. default None

    optimizer: ``Optimizer`` of ``str``
        The (name of) optimizer used to train and predict. default torch.optim.Adam

    lr: ``float``
        The learning rate of node classification task. default 1e-4

    max_epoch: ``int``
        The max number of epochs in training. default 100

    early_stopping_round: ``int``
        The round of early stop. default 100

    weight_decay: ``float``
        weight decay ratio, default 1e-4

    device: ``torch.device`` or ``str``
        The device where model will be running on.

    init: ``bool``
        If True(False), the model will (not) be initialized.

    feval: (Sequence of) ``Evaluation`` or ``str``
        The evaluation functions, default ``[LogLoss]``
    
    loss: ``str``
        The loss used. Default ``nll_loss``.

    lr_scheduler_type: ``str`` (Optional)
        The lr scheduler type used. Default None.

    """

    space = None

    def __init__(
        self,
        model: Union[Tuple[BaseEncoderMaintainer, BaseDecoderMaintainer], BaseEncoderMaintainer, BaseAutoModel, str] = None,
        num_features=None,
        optimizer=torch.optim.Adam,
        lr=1e-4,
        max_epoch=100,
        early_stopping_round=101,
        weight_decay=1e-4,
        device="auto",
        init=False,
        feval=[Auc],
        loss="binary_cross_entropy_with_logits",
        lr_scheduler_type=None,
        *args,
        **kwargs,
    ):
        if isinstance(model, Tuple):
            encoder, decoder = model
        elif isinstance(model, BaseAutoModel):
            encoder, decoder = model, None
        else:
            encoder, decoder = model, "lp-decoder"
        super().__init__(encoder, decoder, num_features, "auto", device, feval, loss)

        self.opt_received = optimizer
        if isinstance(optimizer, str):
            if optimizer.lower() == "adam": self.optimizer = torch.optim.Adam
            elif optimizer.lower() == "sgd": self.optimizer = torch.optim.SGD
            else: raise ValueError("Currently not support optimizer {}".format(optimizer))
        elif isinstance(optimizer, type) and issubclass(optimizer, torch.optim.Optimizer):
            self.optimizer = optimizer
        else:
            raise ValueError("Currently not support optimizer {}".format(optimizer))

        self.lr_scheduler_type = lr_scheduler_type

        self.lr = lr
        self.max_epoch = max_epoch
        self.early_stopping_round = early_stopping_round
        self.args = args
        self.kwargs = kwargs
        self.weight_decay = weight_decay

        self.early_stopping = EarlyStopping(
            patience=early_stopping_round, verbose=False
        )

        self.valid_result = None
        self.valid_result_prob = None
        self.valid_score = None

        self.pyg_dgl = DependentBackend.get_backend_name()

        self.hyper_parameter_space = [
            {
                "parameterName": "max_epoch",
                "type": "INTEGER",
                "maxValue": 500,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "early_stopping_round",
                "type": "INTEGER",
                "maxValue": 30,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "lr",
                "type": "DOUBLE",
                "maxValue": 1e-1,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
            {
                "parameterName": "weight_decay",
                "type": "DOUBLE",
                "maxValue": 1e-2,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
        ]

        self.hyper_parameters = {
            "max_epoch": self.max_epoch,
            "early_stopping_round": self.early_stopping_round,
            "lr": self.lr,
            "weight_decay": self.weight_decay,
        }

        if init is True:
            self.initialize()

    def _compose_model(self):
        return _DummyLinkModel(self.encoder, self.decoder).to(self.device)

    @classmethod
    def get_task_name(cls):
        return "LinkPrediction"

    def _train_only_pyg(self, data, train_mask=None):

        model = self._compose_model()
        data = data.to(self.device)
        optimizer = self.optimizer(
            model.parameters(), lr=self.lr, weight_decay=self.weight_decay
        )
        
        lr_scheduler_type = self.lr_scheduler_type
        if type(lr_scheduler_type) == str and lr_scheduler_type == "steplr":
            scheduler = StepLR(optimizer, step_size=100, gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "multisteplr":
            scheduler = MultiStepLR(optimizer, milestones=[30, 80], gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "exponentiallr":
            scheduler = ExponentialLR(optimizer, gamma=0.1)
        elif (
            type(lr_scheduler_type) == str and lr_scheduler_type == "reducelronplateau"
        ):
            scheduler = ReduceLROnPlateau(optimizer, "min")
        else:
            scheduler = None
        
        for epoch in range(1, self.max_epoch + 1):
            model.train()

            try:
                neg_edge_index = data.train_neg_edge_index
            except:
                from torch_geometric.utils import negative_sampling
                # from ...datasets.utils import negative_sampling

                neg_edge_index = negative_sampling(
                    edge_index=data.train_pos_edge_index,
                    num_nodes=data.num_nodes,
                    num_neg_samples=data.train_pos_edge_index.size(1),
                ).to(data.train_pos_edge_index.device)
            
            optimizer.zero_grad()
            link_logits = model.encode(data)
            link_logits = model.decode(link_logits, data, data.train_pos_edge_index, neg_edge_index)
            link_labels = self.get_link_labels(
                data.train_pos_edge_index, neg_edge_index
            )
            
            if hasattr(F, self.loss):
                loss = getattr(F, self.loss)(link_logits, link_labels)
            else:
                raise TypeError(
                    "PyTorch does not support loss type {}".format(self.loss)
                )

            loss.backward()
            optimizer.step()
            if scheduler:
                scheduler.step()

            if type(self.feval) is list:
                feval = self.feval[0]
            else:
                feval = self.feval
            val_loss = self.evaluate([data], mask="val", feval=feval)
            if feval.is_higher_better() is True:
                val_loss = -val_loss
            self.early_stopping(val_loss, model)
            if self.early_stopping.early_stop:
                LOGGER.debug("Early stopping at %d", epoch)
                break
        self.early_stopping.load_checkpoint(model)

    def _train_only_dgl(self, data):
        model = self._compose_model()
        train_graph = data['train'].to(self.device)
        train_pos_graph = data['train_pos'].to(self.device)
        train_neg_data = data['train_neg'].to(self.device)

        optimizer = self.optimizer(
            model.parameters(), lr=self.lr, weight_decay=self.weight_decay
        )
        
        lr_scheduler_type = self.lr_scheduler_type
        if type(lr_scheduler_type) == str and lr_scheduler_type == "steplr":
            scheduler = StepLR(optimizer, step_size=100, gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "multisteplr":
            scheduler = MultiStepLR(optimizer, milestones=[30, 80], gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "exponentiallr":
            scheduler = ExponentialLR(optimizer, gamma=0.1)
        elif (
            type(lr_scheduler_type) == str and lr_scheduler_type == "reducelronplateau"
        ):
            scheduler = ReduceLROnPlateau(optimizer, "min")
        else:
            scheduler = None

        for epoch in range(1, self.max_epoch + 1):
            model.train()

            optimizer.zero_grad()

            pos_edges, neg_edges = torch.stack(train_pos_graph.edges()), torch.stack(train_neg_data.edges())

            link_logits = model.encode(train_graph)
            link_logits = model.decode(link_logits, train_graph, pos_edges, neg_edges)
            link_labels = self.get_link_labels(pos_edges, neg_edges)
            if hasattr(F, self.loss):
                loss = getattr(F, self.loss)(link_logits, link_labels)
            else:
                raise TypeError(
                    "PyTorch does not support loss type {}".format(self.loss)
                )

            loss.backward()
            optimizer.step()
            if self.lr_scheduler_type:
                scheduler.step()

            if type(self.feval) is list:
                feval = self.feval[0]
            else:
                feval = self.feval
            val_loss = self._evaluate_dgl([data], mask="val", feval=feval)
            if feval.is_higher_better() is True:
                val_loss = -val_loss
            self.early_stopping(val_loss, model)
            if self.early_stopping.early_stop:
                LOGGER.debug("Early stopping at %d", epoch)
                break

        self.early_stopping.load_checkpoint(model)

    def _predict_only_pyg(self, data):
        data = data.to(self.device)
        model = self._compose_model()
        model.eval()
        with torch.no_grad():
            res = model.encode(data)
        return res

    def _predict_only_dgl(self, dataset):
        pos_data = dataset['train'].to(self.device)
        model = self._compose_model()
        model.eval()
        with torch.no_grad():
            z = model.encode(pos_data)
        return z

    def train(self, dataset, keep_valid_result=True):
        """
        train on the given dataset

        Parameters
        ----------
        dataset: The link prediction dataset used to be trained.

        keep_valid_result: ``bool``
            If True(False), save the validation result after training.

        Returns
        -------
        self: ``autogl.train.LinkPredictionTrainer``
            A reference of current trainer.

        """
        if self.pyg_dgl == 'pyg':
            data = dataset[0]
            data.edge_index = data.train_pos_edge_index
            self._train_only_pyg(data)
            if keep_valid_result:
                self.valid_result = self._predict_only_pyg(data)
                self.valid_result_prob = self._predict_proba_pyg(dataset, "val")
                self.valid_score = self._evaluate_pyg(dataset, mask="val", feval=self.feval)
        elif self.pyg_dgl == 'dgl':
            data = dataset[0]
            self._train_only_dgl(data)
            if keep_valid_result:
                self.valid_result = self._predict_only_dgl(data)
                self.valid_result_prob = self._predict_proba_dgl(dataset, "val")
                self.valid_score = self._evaluate_dgl(dataset, mask="val", feval=self.feval)

    def predict(self, dataset, mask=None):
        """
        The function of predicting on the given dataset.

        Parameters
        ----------
        dataset: The link prediction dataset used to be predicted.

        mask: ``train``, ``val``, or ``test``.
            The dataset mask.
        
        .. Note:: Deprecated, this function will be removed in the future.

        Returns
        -------
        The prediction result of ``predict_proba``.
        """
        if self.pyg_dgl == 'pyg':
            return self._predict_proba_pyg(dataset, mask=mask, in_log_format=False)
        elif self.pyg_dgl == 'dgl':
            return self._predict_proba_dgl(dataset, mask=mask, in_log_format=False)

    def predict_proba(self, dataset, mask=None, in_log_format=False):
        if self.pyg_dgl == 'pyg':
            return self._predict_proba_pyg(dataset, mask, in_log_format)
        elif self.pyg_dgl == 'dgl':
            return self._predict_proba_dgl(dataset, mask, in_log_format)

    def _predict_proba_pyg(self, dataset, mask=None, in_log_format=False):
        data = dataset[0]
        data.edge_index = data.train_pos_edge_index
        data = data.to(self.device)
        try:
            if mask in ["train", "val", "test"]:
                pos_edge_index = data[f"{mask}_pos_edge_index"]
                neg_edge_index = data[f"{mask}_neg_edge_index"]
            else:
                pos_edge_index = data[f"test_pos_edge_index"]
                neg_edge_index = data[f"test_neg_edge_index"]
        except:
            pos_edge_index = data[f"test_edge_index"]
            neg_edge_index = torch.zeros(2, 0).to(self.device)

        model = self._compose_model()
        model.eval()
        with torch.no_grad():
            z = self._predict_only_pyg(data)
            link_logits = model.decode(z, data, pos_edge_index, neg_edge_index)
            link_probs = link_logits.sigmoid()

        return link_probs

    def _predict_proba_dgl(self, dataset, mask=None, in_log_format=False):
        dataset = dataset[0]
        train_graph = dataset['train'].to(self.device)
        try:
            try:
                pos_graph = dataset[f'{mask}_pos'].to(self.device)
                neg_graph = dataset[f'{mask}_neg'].to(self.device)
            except:
                pos_graph = dataset[f'test_pos'].to(self.device)
                neg_graph = dataset[f'test_neg'].to(self.device)
        except:
            import dgl
            pos_graph = dataset[mask].to(self.device)
            neg_graph = dgl.graph([], num_nodes=0).to(self.device)

        model = self._compose_model()
        model.eval()
        with torch.no_grad():
            z = model.encode(train_graph)
            link_logits = model.decode(
                z, 
                train_graph,
                torch.stack(pos_graph.edges()), 
                torch.stack(neg_graph.edges())
            )
            link_probs = link_logits.sigmoid()

        return link_probs

    def get_valid_predict(self):
        return self.valid_result

    def get_valid_predict_proba(self):
        return self.valid_result_prob

    def get_valid_score(self, return_major=True):
        """
        The function of getting the valid score.

        Parameters
        ----------
        return_major: ``bool``.
            If True, the return only consists of the major result.
            If False, the return consists of the all results.

        Returns
        -------
        result: The valid score in training stage.
        """
        if isinstance(self.feval, list):
            if return_major:
                return self.valid_score[0], self.feval[0].is_higher_better()
            else:
                return self.valid_score, [f.is_higher_better() for f in self.feval]
        else:
            return self.valid_score, self.feval.is_higher_better()

    def __repr__(self) -> str:
        import yaml

        return yaml.dump(
            {
                "trainer_name": self.__class__.__name__,
                "optimizer": self.optimizer,
                "learning_rate": self.lr,
                "max_epoch": self.max_epoch,
                "early_stopping_round": self.early_stopping_round,
                "encoder": repr(self.encoder),
                "decoder": repr(self.decoder)
            }
        )

    def evaluate(self, dataset, mask=None, feval=None):
        """
        The function of training on the given dataset and keeping valid result.

        Parameters
        ----------
        dataset: The link prediction dataset used to be evaluated.

        mask: ``train``, ``val``, or ``test``.
            The dataset mask.

        feval: ``str``.
            The evaluation method used in this function.

        Returns
        -------
        res: The evaluation result on the given dataset.

        """
        if self.pyg_dgl == 'pyg':
            return self._evaluate_pyg(dataset, mask, feval)
        elif self.pyg_dgl == 'dgl':
            return self._evaluate_dgl(dataset,mask,feval)

    def _evaluate_pyg(self, dataset, mask=None, feval=None):
        data = dataset[0]
        data = data.to(self.device)
        if feval is None:
            feval = self.feval
        else:
            feval = get_feval(feval)

        if mask in ["train", "val", "test"]:
            pos_edge_index = data[f"{mask}_pos_edge_index"]
            neg_edge_index = data[f"{mask}_neg_edge_index"]
        else:
            pos_edge_index = data[f"test_pos_edge_index"]
            neg_edge_index = data[f"test_neg_edge_index"]

        model = self._compose_model()
        model.eval()
        with torch.no_grad():
            link_probs = self._predict_proba_pyg(dataset, mask)
            link_labels = self.get_link_labels(pos_edge_index, neg_edge_index)

        if not isinstance(feval, list):
            feval = [feval]
            return_signle = True
        else:
            return_signle = False

        res = []
        for f in feval:
            res.append(f.evaluate(link_probs.cpu().numpy(), link_labels.cpu().numpy()))
        if return_signle:
            return res[0]
        return res


    def _evaluate_dgl(self, dataset, mask=None, feval=None):
        dataset = dataset[0]
        if feval is None:
            feval = self.feval
        else:
            feval = get_feval(feval)

        train_graph = dataset['train'].to(self.device)
        try:
            pos_graph = dataset[f'{mask}_pos'].to(self.device)
            neg_graph = dataset[f'{mask}_neg'].to(self.device)
        except:
            pos_graph = dataset[f'test_pos'].to(self.device)
            neg_graph = dataset[f'test_neg'].to(self.device)

        model = self._compose_model()
        model.eval()
        with torch.no_grad():
            z = model.encode(train_graph)
            link_logits = model.decode(
                    z,
                    train_graph,
                    torch.stack(pos_graph.edges()),
                    torch.stack(neg_graph.edges())
                )
            link_probs = link_logits.sigmoid()
            link_labels = self.get_link_labels(
                torch.stack(pos_graph.edges()), torch.stack(neg_graph.edges())
            )
        if not isinstance(feval, list):
            feval = [feval]
            return_signle = True
        else:
            return_signle = False

        res = []
        for f in feval:
            res.append(f.evaluate(link_probs.cpu().numpy(), link_labels.cpu().numpy()))
        if return_signle:
            return res[0]
        return res

    def duplicate_from_hyper_parameter(self, hp: dict, model=None, restricted=True):
        """
        The function of duplicating a new instance from the given hyperparameter.

        Parameters
        ----------
        hp: ``dict``.
            The hyperparameter used in the new instance. Should contain 3 keys "trainer", "encoder"
            "decoder", with corresponding hyperparameters as values.

        model: The new model
            Models can be ``str``, ``autogl.module.model.BaseAutoModel``, 
            ``autogl.module.model.encoders.BaseEncoderMaintainer`` or a tuple of (encoder, decoder) 
            if need to specify both encoder and decoder. Encoder can be ``str`` or
            ``autogl.module.model.encoders.BaseEncoderMaintainer``, and decoder can be ``str``
            or ``autogl.module.model.decoders.BaseDecoderMaintainer``.

        restricted: ``bool``.
            If False(True), the hyperparameter should (not) be updated from origin hyperparameter.

        Returns
        -------
        self: ``autogl.train.LinkPredictionTrainer``
            A new instance of trainer.

        """
        if isinstance(model, Tuple):
            encoder, decoder = model
        elif isinstance(model, BaseAutoModel):
            encoder, decoder = model, None
        elif isinstance(model, BaseEncoderMaintainer):
            encoder, decoder = model, self.decoder
        elif model is None:
            encoder, decoder = self.encoder, self.decoder
        else:
            raise TypeError("Cannot parse model with type", type(model))
        
        trainer_hp = hp.get("trainer", {})
        encoder_hp = hp.get("encoder", {})
        decoder_hp = hp.get("decoder", {})

        if not restricted:
            origin_hp = deepcopy(self.hyper_parameters)
            origin_hp.update(trainer_hp)
            trainer_hp = origin_hp
        
        encoder = encoder.from_hyper_parameter(encoder_hp)
        if decoder is not None:
            decoder = decoder.from_hyper_parameter_and_encoder(decoder_hp, encoder)
        
        ret = self.__class__(
            model=(encoder, decoder),
            num_features=self.num_features,
            optimizer=self.optimizer,
            lr=trainer_hp["lr"],
            max_epoch=trainer_hp["max_epoch"],
            early_stopping_round=trainer_hp["early_stopping_round"],
            device=self.device,
            weight_decay=trainer_hp["weight_decay"],
            feval=self.feval,
            init=True,
            *self.args,
            **self.kwargs,
        )

        return ret

    def get_link_labels(self, pos_edge_index, neg_edge_index):
        E = pos_edge_index.size(1) + neg_edge_index.size(1)
        link_labels = torch.zeros(E, dtype=torch.float, device=self.device)
        link_labels[: pos_edge_index.size(1)] = 1.0
        return link_labels



from .. import register_feature
import lightgbm as lgb
from sklearn.model_selection import train_test_split
import psutil
from .base import BaseSelector
import numpy as np
import copy
import pandas as pd


def gbdt_gen(
    data,
    fixlen=1000,
    params={
        "boosting_type": "gbdt",
        "verbosity": -1,
        "random_state": 47,
        "objective": "multiclass",
        "metric": ["multi_logloss"],
        "max_bin": 63,
        "save_binary": True,
        "num_threads": 20,
        "num_leaves": 16,
        "subsample": 0.9,
        "subsample_freq": 1,
        "colsample_bytree": 0.8,
        #         'is_training_metric': True,
        #         'metric_freq': 1,
    },
    is_val=True,
    train_val_ratio=0.2,
    **param_o
):

    n_class = np.max(data.y) + 1
    params["num_class"] = n_class
    #     cpu_count = psutil.cpu_count()
    #     params['num_threads'] = cpu_count if data.train_ind.shape[0] > 10000 else min(32, cpu_count)

    param = {"num_boost_round": 100, "early_stopping_rounds": 5, "verbose_eval": False}
    param.update(param_o)
    if hasattr(data, "train_mask"):
        x = data.x[data.train_mask]
        label = data.y[data.train_mask]
    else:
        x = data.x
        label = data.y
        is_val = False
    _, num_feas = x.shape
    if num_feas < fixlen:
        return None

    fnames = np.array(["f{}".format(i) for i in range(num_feas)])

    if is_val:
        x_train, x_val, y_train, y_val = train_test_split(
            x, label, test_size=train_val_ratio, stratify=label, random_state=47
        )
        dtrain = lgb.Dataset(x_train, label=y_train)
        dval = lgb.Dataset(x_val, label=y_val)
        clf = lgb.train(train_set=dtrain, params=params, valid_sets=dval, **param)
    else:
        train_x = pd.DataFrame(x, columns=fnames, index=None)
        dtrain = lgb.Dataset(train_x, label=label)
        clf = lgb.train(train_set=dtrain, params=params, **param)

    imp = np.array(list(clf.feature_importance()))
    return np.argsort(imp)[-fixlen:]


@register_feature("gbdt")
class SeGBDT(BaseSelector):
    r"""simple wrapper of lightgbm , using importance ranking to select top-k features.

    Parameters
    ----------
    fixlen : int
        K for top-K important features.
    """

    def __init__(self, fixlen=10, *args, **kwargs):
        super(SeGBDT, self).__init__()
        self.fixlen = fixlen
        self.args = args
        self.kwargs = kwargs

    def _fit(self, data):
        self._sel = gbdt_gen(data, self.fixlen, *self.args, **self.kwargs)



from ._dgl_decoders import (
    LogSoftmaxDecoderMaintainer,
    JKSumPoolDecoderMaintainer,
    SumPoolMLPDecoderMaintainer,
    TopKDecoderMaintainer,
    DotProductLinkPredictionDecoderMaintainer
)



import typing as _typing
from autogl.utils import universal_registry
from . import base_encoder


class EncoderUniversalRegistry(universal_registry.UniversalRegistryBase):
    @classmethod
    def register_encoder(cls, name: str) -> _typing.Callable[
        [_typing.Type[base_encoder.BaseEncoderMaintainer]],
        _typing.Type[base_encoder.BaseEncoderMaintainer]
    ]:
        def register_encoder(
                _encoder: _typing.Type[base_encoder.BaseEncoderMaintainer]
        ) -> _typing.Type[base_encoder.BaseEncoderMaintainer]:
            if not issubclass(_encoder, base_encoder.BaseEncoderMaintainer):
                raise TypeError
            else:
                cls[name] = _encoder
                return _encoder

        return register_encoder

    @classmethod
    def get_encoder(cls, name: str) -> _typing.Type[base_encoder.BaseEncoderMaintainer]:
        if name not in cls:
            raise ValueError(f"Encoder with name \"{name}\" not exist")
        else:
            return cls[name]



from ...backend import DependentBackend

TRAINER_DICT = {}
from .base import (
    BaseTrainer,
    Evaluation,
    BaseNodeClassificationTrainer,
    BaseGraphClassificationTrainer,
    BaseLinkPredictionTrainer,
    BaseNodeClassificationHetTrainer
)


def register_trainer(name):
    def register_trainer_cls(cls):
        if name in TRAINER_DICT:
            raise ValueError("Cannot register duplicate trainer ({})".format(name))
        if not issubclass(cls, BaseTrainer):
            raise ValueError(
                "Trainer ({}: {}) must extend BaseTrainer".format(name, cls.__name__)
            )
        TRAINER_DICT[name] = cls
        return cls

    return register_trainer_cls


from .graph_classification_full import GraphClassificationFullTrainer
from .node_classification_full import NodeClassificationFullTrainer
from .link_prediction_full import LinkPredictionTrainer
from .node_classification_het import NodeClassificationHetTrainer
if DependentBackend.is_pyg():
    from .node_classification_trainer import (
        NodeClassificationGraphSAINTTrainer,
        NodeClassificationLayerDependentImportanceSamplingTrainer,
        NodeClassificationNeighborSamplingTrainer
    )
    from .ssl import GraphCLSemisupervisedTrainer, GraphCLUnsupervisedTrainer
from .evaluation import get_feval, Acc, Auc, Logloss, Mrr, MicroF1

__all__ = [
    "BaseTrainer",
    "Evaluation",
    "BaseGraphClassificationTrainer",
    "BaseNodeClassificationTrainer",
    "BaseNodeClassificationHetTrainer",
    "BaseLinkPredictionTrainer",
    "GraphClassificationFullTrainer",
    "NodeClassificationFullTrainer",
    "NodeClassificationHetTrainer",
    "LinkPredictionTrainer",
    "Acc",
    "Auc",
    "Logloss",
    "Mrr",
    "MicroF1",
    "get_feval",
]

if DependentBackend.is_pyg():
    __all__.extend([
        "NodeClassificationGraphSAINTTrainer",
        "NodeClassificationLayerDependentImportanceSamplingTrainer",
        "NodeClassificationNeighborSamplingTrainer",
        "GraphCLSemisupervisedTrainer",
        "GraphCLUnsupervisedTrainer"
    ])


import math
import torch
import typing as _typing


def _maybe_num_nodes(edge_index, num_nodes=None):
    if isinstance(num_nodes, int):
        return num_nodes
    elif isinstance(edge_index, torch.Tensor):
        return int(edge_index.max()) + 1 if edge_index.numel() > 0 else 0
    else:
        return max(edge_index.size(0), edge_index.size(1))


def __coalesce(
        edge_index: torch.Tensor,
        edge_attr: _typing.Union[
            torch.Tensor, _typing.Iterable[torch.Tensor], None
        ] = None,
        num_nodes: _typing.Optional[int] = ...,
        is_sorted: bool = False,
        sort_by_row: bool = True
) -> _typing.Union[
    torch.Tensor, _typing.Tuple[torch.Tensor, torch.Tensor],
    _typing.Tuple[torch.Tensor, _typing.Iterable[torch.Tensor]]
]:
    """
    Row-wise sorts :obj:`edge_index` and removes its duplicated entries.
    Duplicate entries in :obj:`edge_attr` are directly removed, instead of merged.
    Args:
        edge_index (LongTensor): The edge indices.
        edge_attr (Tensor or List[Tensor], optional): Edge weights or multi-
            dimensional edge features.
            If given as a list, will re-shuffle and remove duplicates for all
            its entries. (default: :obj:`None`)
        num_nodes (int, optional): The number of nodes, *i.e.*
            :obj:`max_val + 1` of :attr:`edge_index`. (default: :obj:`None`)
        is_sorted (bool, optional): If set to :obj:`True`, will expect
            :obj:`edge_index` to be already sorted row-wise.
        sort_by_row (bool, optional): If set to :obj:`False`, will sort
            :obj:`edge_index` column-wise.
    :rtype: :class:`LongTensor` if :attr:`edge_attr` is :obj:`None`, else
        (:class:`LongTensor`, :obj:`Tensor` or :obj:`Iterable[Tensor]]`)
    """

    if edge_attr is None:
        pass
    elif isinstance(edge_attr, torch.Tensor) and torch.is_tensor(edge_attr):
        if edge_attr.size(0) != edge_index.size(1):
            raise ValueError
    elif isinstance(edge_attr, _typing.Iterable):
        if not all([
            (
                    isinstance(attr, torch.Tensor) and
                    attr.size(0) == edge_index.size(1)
            ) for attr in edge_attr
        ]):
            raise ValueError("Invalid edge_attr argument")
    else:
        raise TypeError("Unsupported type of edge_attr argument")

    nnz = edge_index.size(1)
    num_nodes = _maybe_num_nodes(edge_index, num_nodes)

    idx = edge_index.new_empty(nnz + 1)
    idx[0] = -1
    idx[1:] = edge_index[1 - int(sort_by_row)]
    idx[1:].mul_(num_nodes).add_(edge_index[int(sort_by_row)])

    if not is_sorted:
        idx[1:], perm = idx[1:].sort()
        edge_index = edge_index[:, perm]
        if edge_attr is not None and isinstance(edge_attr, torch.Tensor):
            edge_attr = edge_attr[perm]
        elif edge_attr is not None:
            edge_attr = [e[perm] for e in edge_attr]

    mask: _typing.Any = idx[1:] > idx[:-1]

    # Only perform expensive merging in case there exists duplicates:
    if mask.all():
        return edge_index if edge_attr is None else (edge_index, edge_attr)

    edge_index = edge_index[:, mask]
    if edge_attr is None:
        return edge_index
    elif isinstance(edge_attr, torch.Tensor):
        return edge_index, edge_attr[mask]
    elif isinstance(edge_attr, _typing.Iterable):
        return edge_index, [attr[mask] for attr in edge_attr]


def coalesce(
        edge_index: torch.Tensor,
        edge_attr: _typing.Union[
            torch.Tensor, _typing.Iterable[torch.Tensor], None
        ] = None,
        num_nodes: _typing.Optional[int] = ...,
        is_sorted: bool = False,
        sort_by_row: bool = True
) -> _typing.Union[
    torch.Tensor, _typing.Tuple[torch.Tensor, torch.Tensor],
    _typing.Tuple[torch.Tensor, _typing.Iterable[torch.Tensor]]
]:
    """
    Row-wise sorts :obj:`edge_index` and removes its duplicated entries.
    Duplicate entries in :obj:`edge_attr` are directly removed, instead of merged.
    Args:
        edge_index (LongTensor): The edge indices.
        edge_attr (Tensor or List[Tensor], optional): Edge weights or multi-
            dimensional edge features.
            If given as a list, will re-shuffle and remove duplicates for all
            its entries. (default: :obj:`None`)
        num_nodes (int, optional): The number of nodes, *i.e.*
            :obj:`max_val + 1` of :attr:`edge_index`. (default: :obj:`None`)
        is_sorted (bool, optional): If set to :obj:`True`, will expect
            :obj:`edge_index` to be already sorted row-wise.
        sort_by_row (bool, optional): If set to :obj:`False`, will sort
            :obj:`edge_index` column-wise.
    :rtype: :class:`LongTensor` if :attr:`edge_attr` is :obj:`None`, else
        (:class:`LongTensor`, :obj:`Tensor` or :obj:`Iterable[Tensor]]`)
    """
    if not isinstance(num_nodes, int):
        num_nodes = None
    try:
        import torch_geometric
        return torch_geometric.utils.coalesce(
            edge_index, edge_attr, num_nodes,
            is_sorted=is_sorted,
            sort_by_row=sort_by_row
        )
    except ModuleNotFoundError:
        return __coalesce(
            edge_index, edge_attr, num_nodes,
            is_sorted=is_sorted,
            sort_by_row=sort_by_row
        )


def to_undirected(
        edge_index: torch.Tensor,
        edge_attr: _typing.Optional[_typing.Union[torch.Tensor, _typing.List[torch.Tensor]]] = None,
        num_nodes: _typing.Optional[int] = ...,
        __reduce: str = "add",
) -> _typing.Union[
    torch.Tensor, _typing.Tuple[torch.Tensor, torch.Tensor],
    _typing.Tuple[torch.Tensor, _typing.List[torch.Tensor]]
]:
    r"""Converts the graph given by :attr:`edge_index` to an undirected graph
    such that :math:`(j,i) \in \mathcal{E}` for every edge :math:`(i,j) \in
    \mathcal{E}`.

    Args:
        edge_index (LongTensor): The edge indices.
        edge_attr (Tensor or List[Tensor], optional): Edge weights or multi-
            dimensional edge features.
            If given as a list, will remove duplicates for all its entries.
            (default: :obj:`None`)
        num_nodes (int, optional): The number of nodes, *i.e.*
            :obj:`max_val + 1` of :attr:`edge_index`. (default: :obj:`None`)
        __reduce (string, optional): The reduce operation to use for merging edge
            features (:obj:`"add"`, :obj:`"mean"`, :obj:`"min"`, :obj:`"max"`,
            :obj:`"mul"`). (default: :obj:`"add"`)

    :rtype: :class:`LongTensor` if :attr:`edge_attr` is :obj:`None`, else
        (:class:`LongTensor`, :obj:`Tensor` or :obj:`List[Tensor]]`)
    """
    # Maintain backward compatibility to `to_undirected(edge_index, num_nodes)`
    if isinstance(edge_attr, int):
        edge_attr = None
        num_nodes = edge_attr

    row, col = edge_index
    row, col = torch.cat([row, col], dim=0), torch.cat([col, row], dim=0)
    edge_index = torch.stack([row, col], dim=0)

    if edge_attr is not None and isinstance(edge_attr, torch.Tensor):
        edge_attr = torch.cat([edge_attr, edge_attr], dim=0)
    elif edge_attr is not None:
        edge_attr = [torch.cat([e, e], dim=0) for e in edge_attr]

    return coalesce(edge_index, edge_attr, num_nodes)


class _SplitResult:
    def __init__(
            self,
            train_pos_edge_index: torch.Tensor,
            train_pos_edge_attr: _typing.Optional[torch.Tensor],
            train_neg_adj_mask: torch.Tensor,
            val_pos_edge_index: torch.Tensor,
            val_pos_edge_attr: _typing.Optional[torch.Tensor],
            val_neg_edge_index: torch.Tensor,
            test_pos_edge_index: torch.Tensor,
            test_pos_edge_attr: _typing.Optional[torch.Tensor],
            test_neg_edge_index: torch.Tensor
    ):
        self.train_pos_edge_index: torch.Tensor = train_pos_edge_index
        self.train_pos_edge_attr: _typing.Optional[torch.Tensor] = train_pos_edge_attr
        self.train_neg_adj_mask: torch.Tensor = train_neg_adj_mask
        self.val_pos_edge_index: torch.Tensor = val_pos_edge_index
        self.val_pos_edge_attr: _typing.Optional[torch.Tensor] = val_pos_edge_attr
        self.val_neg_edge_index: torch.Tensor = val_neg_edge_index
        self.test_pos_edge_index: torch.Tensor = test_pos_edge_index
        self.test_pos_edge_attr: _typing.Optional[torch.Tensor] = test_pos_edge_attr
        self.test_neg_edge_index: torch.Tensor = test_neg_edge_index


def train_test_split_edges(
        edge_index: torch.Tensor,
        edge_attr: _typing.Optional[_typing.Union[torch.Tensor, _typing.List[torch.Tensor]]] = None,
        num_nodes: _typing.Optional[int] = ...,
        val_ratio: float = 0.05,
        test_ratio: float = 0.1
):
    r"""Splits the edges of a :class:`torch_geometric.data.Data` object
    into positive and negative train/val/test edges.
    As such, it will replace the :obj:`edge_index` attribute with
    :obj:`train_pos_edge_index`, :obj:`train_pos_neg_adj_mask`,
    :obj:`val_pos_edge_index`, :obj:`val_neg_edge_index` and
    :obj:`test_pos_edge_index` attributes.
    If :obj:`data` has edge features named :obj:`edge_attr`, then
    :obj:`train_pos_edge_attr`, :obj:`val_pos_edge_attr` and
    :obj:`test_pos_edge_attr` will be added as well.

    .. warning::

        :meth:`~torch_geometric.utils.train_test_split_edges` is deprecated and
        will be removed in a future release.
        Use :class:`torch_geometric.transforms.RandomLinkSplit` instead.

    Args:
        edge_index (LongTensor): The edge indices.
        edge_attr (Tensor or List[Tensor], optional): Edge weights or multi-
            dimensional edge features.
            If given as a list, will remove duplicates for all its entries.
            (default: :obj:`None`)
        num_nodes (int, optional): The number of nodes, *i.e.*
            :obj:`max_val + 1` of :attr:`edge_index`. (default: :obj:`None`)
        val_ratio (float, optional): The ratio of positive validation edges.
            (default: :obj:`0.05`)
        test_ratio (float, optional): The ratio of positive test edges.
            (default: :obj:`0.1`)
    """
    row, col = edge_index
    num_nodes = _maybe_num_nodes(edge_index, num_nodes)

    # Return upper triangular portion.
    mask = row < col
    row, col = row[mask], col[mask]

    if edge_attr is not None:
        edge_attr = edge_attr[mask]

    n_v = int(math.floor(val_ratio * row.size(0)))
    n_t = int(math.floor(test_ratio * row.size(0)))

    # Positive edges.
    perm = torch.randperm(row.size(0))
    row, col = row[perm], col[perm]
    if edge_attr is not None:
        edge_attr = edge_attr[perm]

    r, c = row[:n_v], col[:n_v]
    val_pos_edge_index = torch.stack([r, c], dim=0)
    if edge_attr is not None:
        val_pos_edge_attr = edge_attr[:n_v]
    else:
        val_pos_edge_attr = None

    r, c = row[n_v:n_v + n_t], col[n_v:n_v + n_t]
    test_pos_edge_index = torch.stack([r, c], dim=0)
    if edge_attr is not None:
        test_pos_edge_attr = edge_attr[n_v:n_v + n_t]
    else:
        test_pos_edge_attr = None

    r, c = row[n_v + n_t:], col[n_v + n_t:]
    train_pos_edge_index = torch.stack([r, c], dim=0)
    if edge_attr is not None:
        train_pos_edge_index, train_pos_edge_attr = to_undirected(
            train_pos_edge_index, edge_attr[n_v + n_t:]
        )
    else:
        train_pos_edge_index = to_undirected(train_pos_edge_index)
        train_pos_edge_attr = None

    # Negative edges.
    neg_adj_mask = torch.ones(num_nodes, num_nodes, dtype=torch.uint8)
    neg_adj_mask = neg_adj_mask.triu(diagonal=1).to(torch.bool)
    neg_adj_mask[row, col] = 0

    neg_row, neg_col = neg_adj_mask.nonzero(as_tuple=False).t()
    perm = torch.randperm(neg_row.size(0))[:n_v + n_t]
    neg_row, neg_col = neg_row[perm], neg_col[perm]

    neg_adj_mask[neg_row, neg_col] = 0
    train_neg_adj_mask = neg_adj_mask

    row, col = neg_row[:n_v], neg_col[:n_v]
    val_neg_edge_index = torch.stack([row, col], dim=0)

    row, col = neg_row[n_v:n_v + n_t], neg_col[n_v:n_v + n_t]
    test_neg_edge_index = torch.stack([row, col], dim=0)

    return _SplitResult(
        train_pos_edge_index, train_pos_edge_attr, train_neg_adj_mask,
        val_pos_edge_index, val_pos_edge_attr, val_neg_edge_index,
        test_pos_edge_index, test_pos_edge_attr, test_neg_edge_index
    )



import copy
import torch
import typing as _typing
from autogl.data.graph import GeneralStaticGraph
from . import _base_feature_engineer


class BaseFeatureEngineer(
    _base_feature_engineer.BaseFeatureEngineer
):
    @classmethod
    def __preprocess(
            cls, data: _typing.Union[GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[GeneralStaticGraph, _typing.Any]:
        return data  # todo: Support torch_geometric.HeteroData in future

    @classmethod
    def __postprocess(
            cls, data: _typing.Union[GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[GeneralStaticGraph, _typing.Any]:
        return data  # todo: Support torch_geometric.HeteroData in future

    def fit(self, dataset):
        dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self.__postprocess(
                    self._postprocess(self._fit(self._preprocess(self.__preprocess(data))))
                )
        return dataset

    def transform(self, dataset, inplace: bool = True):
        if not inplace:
            dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self.__postprocess(
                    self._postprocess(self._transform(self._preprocess(self.__preprocess(data))))
                )
        return dataset



import torch
import torch.nn.functional as F
from torch_geometric.nn import GraphConv, TopKPooling
from torch_geometric.nn import global_mean_pool as gap, global_max_pool as gmp
from . import register_model
from .base import BaseAutoModel, activate_func
from ....utils import get_logger

LOGGER = get_logger("TopkModel")


def set_default(args, d):
    for k, v in d.items():
        if k not in args:
            args[k] = v
    return args


class Topkpool(torch.nn.Module):
    def __init__(self, args):
        super(Topkpool, self).__init__()
        self.args = args

        missing_keys = list(
            set(
                [
                    "features_num",
                    "num_class",
                    "num_graph_features",
                    "ratio",
                    "dropout",
                    "act",
                ]
            )
            - set(self.args.keys())
        )
        if len(missing_keys) > 0:
            raise Exception("Missing keys: %s." % ",".join(missing_keys))

        self.num_features = self.args["features_num"]
        self.num_classes = self.args["num_class"]
        self.ratio = self.args["ratio"]
        self.dropout = self.args["dropout"]
        self.num_graph_features = self.args["num_graph_features"]

        self.conv1 = GraphConv(self.num_features, 128)
        self.pool1 = TopKPooling(128, ratio=self.ratio)
        self.conv2 = GraphConv(128, 128)
        self.pool2 = TopKPooling(128, ratio=self.ratio)
        self.conv3 = GraphConv(128, 128)
        self.pool3 = TopKPooling(128, ratio=self.ratio)

        self.lin1 = torch.nn.Linear(256 + self.num_graph_features, 128)
        self.lin2 = torch.nn.Linear(128, 64)
        self.lin3 = torch.nn.Linear(64, self.num_classes)

    def forward(self, data):
        x, edge_index, batch = data.x, data.edge_index, data.batch
        if self.num_graph_features > 0:
            graph_feature = data.gf

        x = F.relu(self.conv1(x, edge_index))
        x, edge_index, _, batch, _, _ = self.pool1(x, edge_index, None, batch)
        x1 = torch.cat([gmp(x, batch), gap(x, batch)], dim=1)

        x = F.relu(self.conv2(x, edge_index))
        x, edge_index, _, batch, _, _ = self.pool2(x, edge_index, None, batch)
        x2 = torch.cat([gmp(x, batch), gap(x, batch)], dim=1)

        x = F.relu(self.conv3(x, edge_index))
        x, edge_index, _, batch, _, _ = self.pool3(x, edge_index, None, batch)
        x3 = torch.cat([gmp(x, batch), gap(x, batch)], dim=1)

        x = x1 + x2 + x3
        if self.num_graph_features > 0:
            x = torch.cat([x, graph_feature], dim=-1)
        x = self.lin1(x)
        x = activate_func(x, self.args["act"])
        x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.lin2(x)
        x = activate_func(x, self.args["act"])
        x = F.log_softmax(self.lin3(x), dim=-1)

        return x


@register_model("topkpool-model")
class AutoTopkpool(BaseAutoModel):
    r"""
    AutoTopkpool. The model used in this automodel is from https://arxiv.org/abs/1905.05178, https://arxiv.org/abs/1905.02850

    Parameters
    ----------
    num_features: `int`.
        The dimension of features.

    num_classes: `int`.
        The number of classes.

    device: `torch.device` or `str`
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.

    """

    def __init__(
        self,
        num_features=None,
        num_classes=None,
        device=None,
        init=False,
        num_graph_features=0,
        **args
    ):
        super().__init__(num_features, num_classes, device, num_graph_features=num_graph_features, **args)
        self.num_graph_features = num_graph_features
        self.hyper_parameter_space = [
            {
                "parameterName": "ratio",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]

        self.hyper_parameters = {"ratio": 0.8, "dropout": 0.5, "act": "relu"}


    def from_hyper_parameter(self, hp, **kwargs):
        return super().from_hyper_parameter(hp, num_graph_features=self.num_graph_features, **kwargs)

    def _initialize(self):
        self._model = Topkpool({
            "features_num": self.input_dimension,
            "num_class": self.output_dimension,
            "num_graph_features": self.num_graph_features,
            **self.hyper_parameters
        }).to(self.device)



import typing as _typ

from . import register_nas_space
from .base import apply_fixed_architecture
from .base import BaseSpace
from ...model import BaseAutoModel
from ....utils import get_logger

from ..backend import *
from ..utils import count_parameters, measure_latency

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import Module
from .gasso_space import *
from torch.autograd import Variable
from collections import namedtuple

Genotype = namedtuple('Genotype', 'normal normal_concat')
Genotype_normal = namedtuple('Genotype_normal', 'normal normal_concat')

gnn_list = [
    "gat",  # GAT with 2 heads
    "gcn",  # GCN
    "gin",  # GIN
    #"cheb",  # chebnet
    "sage",  # sage
    #"arma",
    #"sg",  # simplifying gcn
    "linear",  # skip connection
    #"skip",  # skip connection
    #"zero",  # skip connection
]
act_list = [
    "sigmoid", "tanh", "relu", "linear", "elu"
]

def gnn_map(gnn_name, in_dim, out_dim, concat=False, bias=True) -> Module:
    '''

    :param gnn_name:
    :param in_dim:
    :param out_dim:
    :param concat: for gat, concat multi-head output or not
    :return: GNN model
    '''
    norm= True
    if gnn_name == "gat":
        return GATConv(in_dim, out_dim, 1, bias=bias, concat = False, add_self_loops=norm)
    elif gnn_name == "gcn":
        return GCNConv(in_dim, out_dim, add_self_loops=True, normalize=norm)
    elif gnn_name == "gin":
        return GINConv(torch.nn.Linear(in_dim, out_dim))
    elif gnn_name == "cheb":
        return ChebConv(in_dim, out_dim, K=2, bias=bias)
    elif gnn_name == "sage":
        return SAGEConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "gated":
        return GatedGraphConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "arma":
        return ARMAConv(in_dim, out_dim, bias=bias, normalize=norm)
    elif gnn_name == "sg":
        return SGConv(in_dim, out_dim, bias=bias, normalize=norm)
    elif gnn_name == "linear":
        return LinearConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "skip":
        return SkipConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "zero":
        return ZeroConv(in_dim, out_dim, bias=bias)
    else:
        raise ValueError("No such GNN name") 

def Get_edges(adjs, ):
    edges = []
    edges_weights = []
    for adj in adjs: 
        edges.append(adj[0])
        edges_weights.append(torch.sigmoid(adj[1]))
    return edges, edges_weights

class LinearConv(Module):
    def __init__(self,
                 in_channels,
                 out_channels,
                 bias=True):
        super(LinearConv, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.linear = torch.nn.Linear(in_channels, out_channels, bias)

    def forward(self, x, edge_index, edge_weight=None):
        return self.linear(x)

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.in_channels,
                                   self.out_channels)

class SkipConv(Module):
    def __init__(self,
                 in_channels,
                 out_channels,
                 bias=True):
        super(SkipConv, self).__init__()
        self.out_dim = out_channels


    def forward(self, x, edge_index, edge_weight=None):
        return x

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.in_channels,
                                   self.out_channels)

class ZeroConv(Module):
    def __init__(self,
                 in_channels,
                 out_channels,
                 bias=True):
        super(ZeroConv, self).__init__()
        self.out_dim = out_channels


    def forward(self, x, edge_index, edge_weight=None):
        return torch.zeros([x.size(0), self.out_dim]).to(x.device)

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.in_channels,
                                   self.out_channels)

class MixedOp(nn.Module):

    def __init__(self, in_c, out_c):
        super(MixedOp, self).__init__()
        self._ops = nn.ModuleList()
        for action in gnn_list:
            self._ops.append(gnn_map(action, in_c, out_c))

    def forward(self, x, edge_index, edge_weight, weights, selected_idx=None):
        if selected_idx is None:
            fin = []
            for w, op, op_name in zip(weights, self._ops, gnn_list):
                """if op_name == "gcn":
                    w = 1.0
                else:
                    continue"""
                if edge_weight == None:
                    fin.append(w * op(x, edge_index))
                else:
                    fin.append(w * op(x, edge_index, edge_weight = edge_weight))
            return sum(fin)
            #return sum(w * op(x, edge_index) for w, op in zip(weights, self._ops))
        else:  # unchosen operations are pruned
            return self._ops[selected_idx](x, edge_index)

class CellWS(nn.Module):

    def __init__(self, steps, his_dim, hidden_dim, out_dim, dp, bias=True):
        super(CellWS, self).__init__()
        self.steps = steps
        self._ops = nn.ModuleList()
        self._bns = nn.ModuleList()
        self.use2 = False
        self.dp = 0.8
        for i in range(self.steps):
            if i == 0: 
                inpdim = his_dim
            else:
                inpdim = hidden_dim
            if i == self.steps - 1:
                oupdim = out_dim
            else:
                oupdim = hidden_dim
            op = MixedOp(inpdim, oupdim)
            self._ops.append(op)
            self._bns.append(nn.BatchNorm1d(oupdim))

    def forward(self, x, adjs, weights):
        edges, ews = Get_edges(adjs)
        for i in range(self.steps):
            if i > 0:
                x = F.relu(x)
                x = F.dropout(x, p=self.dp, training=self.training)
            x = self._ops[i](x, edges[i], ews[i], weights[i])  # call the gcn module
        return x

@register_nas_space("gassospace")
class GassoSpace(BaseSpace):
    def __init__(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.8,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = gnn_list,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.hidden_dim = hidden_dim
        self.steps = layer_number
        self.dropout = dropout
        self.ops = ops
        self.use_forward = True
        self.dead_tensor = torch.nn.Parameter(torch.FloatTensor([1]), requires_grad = True)

    def instantiate(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.8,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = gnn_list,
    ):
        super().instantiate()
        self.input_dim = input_dim or self.input_dim
        self.output_dim = output_dim or self.output_dim
        self.hidden_dim = hidden_dim or self.hidden_dim
        self.steps = layer_number or self.steps
        self.dropout = dropout or self.dropout
        self.ops = ops or self.ops
        his_dim, cur_dim, hidden_dim, out_dim = self.input_dim, self.input_dim, self.hidden_dim, self.hidden_dim
        self.cells = nn.ModuleList()

        self.cell = CellWS(self.steps, his_dim, hidden_dim, self.output_dim, self.dropout)
        his_dim = cur_dim
        cur_dim = self.steps * out_dim

        self.classifier = nn.Linear(cur_dim, self.output_dim)

        self.initialize_alphas()

    #def forward(self, x, adjs):
    def forward(self, data):
        if self.use_forward:
            x, adjs = data.x, data.adj
            x = F.dropout(x, p=self.dropout, training=self.training)

            weights = []
            for j in range(self.steps):
                weights.append(F.softmax(self.alphas_normal[j], dim=-1))

            x = self.cell(x, adjs, weights)
            x = F.log_softmax(x, dim=1)
            self.current_pred = x.detach()
            return x
        else:
            #for i in self.parameters():
            #    print(i)
            x = self.prediction + self.dead_tensor * 0
            return x

    def keep_prediction(self):
        self.prediction = self.current_pred

    '''def to(self, *args, **kwargs):
        fin = super().to(*args, **kwargs)
        device = next(fin.parameters()).device
        fin.alphas_normal = [i.to(device) for i in self.alphas_normal]
        return fin'''

    def initialize_alphas(self):
        num_ops = len(self.ops)

        self.alphas_normal = []
        for i in range(self.steps):
            self.alphas_normal.append(Variable(1e-3 * torch.randn(num_ops), requires_grad=True))

        self._arch_parameters = [
            self.alphas_normal
        ]

    def arch_parameters(self):
        return self.alphas_normal

    def parse_model(self, selection, device) -> BaseAutoModel:
        self.use_forward = False
        return self.wrap()



import numpy as np
import networkx as nx
import torch
import autogl
from ._basic import BaseFeatureGenerator
from ..._data_preprocessor_registry import DataPreprocessorUniversalRegistry


@DataPreprocessorUniversalRegistry.register_data_preprocessor("PageRank".lower())
class PageRankFeatureGenerator(BaseFeatureGenerator):
    def _extract_nodes_feature(self, data: autogl.data.Data) -> torch.Tensor:
        edge_weight = getattr(data, "edge_weight").tolist()
        g = nx.DiGraph()
        g.add_weighted_edges_from(
            [
                (u, v, edge_weight[i])
                for i, (u, v) in enumerate(data.edge_index.t().tolist())
            ]
        )
        page_rank = nx.pagerank(g)
        num_nodes: int = (
            data.x.size(0)
            if data.x is not None and isinstance(data.x, torch.Tensor)
            else (data.edge_index.max().item() + 1)
        )
        pr = np.zeros(num_nodes)
        for i, v in page_rank.items():
            pr[i] = v
        return torch.from_numpy(pr).unsqueeze(-1)



from this import d
import numpy as np
from copy import deepcopy
from numba import njit
import torch.nn as nn
import torch.nn.functional as F
import math
import torch
import torch.optim as optim
#from deeprobust.graph.defense.basicfunction import att_coef
import typing as _typing
from torch_geometric.data import Data
from sklearn.metrics.pairwise import euclidean_distances, cosine_similarity
from scipy.sparse import lil_matrix
from sklearn.preprocessing import normalize
import scipy.sparse as sp

import autogl.data
from .. import register_model
from . import utils
from ..gcn import GCN
#from torch_geometric.nn.conv import GCNConv
from .nn import GCNConv
from ..base import BaseAutoModel
from .....utils import get_logger
from .utils import accuracy

LOGGER = get_logger("GNNGuardModel")


class GCN4GNNGuard(GCN):
    # Based on the existing GCN, add the robust part.
    def __init__(self, nfeat, nclass, nhid, activation, dropout=0.5, lr=0.01, drop=False, weight_decay=5e-4, with_relu=True, with_bias=True, add_self_loops = True, normalize = True):
        super(GCN4GNNGuard, self).__init__(nfeat, nclass, nhid, activation, dropout=dropout, add_self_loops = add_self_loops, normalize = normalize)

        self.lr = lr
        self.weight_decay = weight_decay
        self.dropout = dropout
        self.nclass = nclass
        self.nfeat = nfeat
        self.hidden_sizes = nhid
        self.drop = drop
        if not with_relu:
            self.weight_decay = 0
        else:
            self.weight_decay = weight_decay
        self.with_relu = with_relu
        self.with_bias = with_bias

        self.gc1 = GCNConv(nfeat, nhid[0], bias=True,)
        self.gc2 = GCNConv(nhid[0], nclass, bias=True, )

    def forward(self, x, adj):
        """we don't change the edge_index, just update the edge_weight;
        some edge_weight are regarded as removed if it equals to zero"""
        x = x.to_dense()

        """GCN and GAT"""
        if self.attention:
            adj = self.att_coef(x, adj, i=0)
        # Add this line
        edge_index = adj._indices().to(self.device)

        x = self.gc1(x, edge_index, edge_weight=adj._values())
        x = F.relu(x)
        # x = self.bn1(x)
        if self.attention:  # if attention=True, use attention mechanism
            adj_2 = self.att_coef(x, adj, i=1)
            adj_memory = adj_2.to_dense()  # without memory
            # adj_memory = self.gate * adj.to_dense() + (1 - self.gate) * adj_2.to_dense()
            row, col = adj_memory.nonzero()[:,0], adj_memory.nonzero()[:,1]
            edge_index = torch.stack((row, col), dim=0)
            adj_values = adj_memory[row, col]
        else:
            edge_index = adj._indices()
            adj_values = adj._values()
        # Add this line
        edge_index = edge_index.to(self.device)
        adj_values = adj_values.to(self.device)

        x = F.dropout(x, self.dropout, training=self.training)
        x = self.gc2(x, edge_index, edge_weight=adj_values)
        
        return F.log_softmax(x, dim=1)

    class myData:
        def __init__(self, x, edge_index, edge_weight=None):
            self.x = x
            self.edge_index = edge_index
            self.edge_weight = edge_weight

    def att_coef(self, fea, edge_index, is_lil=False, i=0):
        if is_lil == False:
            edge_index = edge_index._indices()
        else:
            edge_index = edge_index.tocoo()

        n_node = fea.shape[0]
        row, col = edge_index[0].cpu().data.numpy()[:], edge_index[1].cpu().data.numpy()[:]

        fea_copy = fea.cpu().data.numpy()
        sim_matrix = cosine_similarity(X=fea_copy, Y=fea_copy)  # try cosine similarity
        sim = sim_matrix[row, col]
        sim[sim<0.1] = 0
        # print('dropped {} edges'.format(1-sim.nonzero()[0].shape[0]/len(sim)))

        # """use jaccard for binary features and cosine for numeric features"""
        # fea_start, fea_end = fea[edge_index[0]], fea[edge_index[1]]
        # isbinray = np.array_equal(fea_copy, fea_copy.astype(bool))  # check is the fea are binary
        # np.seterr(divide='ignore', invalid='ignore')
        # if isbinray:
        #     fea_start, fea_end = fea_start.T, fea_end.T
        #     sim = jaccard_score(fea_start, fea_end, average=None)  # similarity scores of each edge
        # else:
        #     fea_copy[np.isinf(fea_copy)] = 0
        #     fea_copy[np.isnan(fea_copy)] = 0
        #     sim_matrix = cosine_similarity(X=fea_copy, Y=fea_copy)  # try cosine similarity
        #     sim = sim_matrix[edge_index[0], edge_index[1]]
        #     sim[sim < 0.01] = 0

        """build a attention matrix"""
        att_dense = lil_matrix((n_node, n_node), dtype=np.float32)
        att_dense[row, col] = sim
        if att_dense[0, 0] == 1:
            att_dense = att_dense - sp.diags(att_dense.diagonal(), offsets=0, format="lil")
        # normalization, make the sum of each row is 1
        att_dense_norm = normalize(att_dense, axis=1, norm='l1')


        """add learnable dropout, make character vector"""
        if self.drop:
            character = np.vstack((att_dense_norm[row, col].A1,
                                     att_dense_norm[col, row].A1))
            character = torch.from_numpy(character.T)
            drop_score = self.drop_learn_1(character)
            drop_score = torch.sigmoid(drop_score)  # do not use softmax since we only have one element
            mm = torch.nn.Threshold(0.5, 0)
            drop_score = mm(drop_score)
            mm_2 = torch.nn.Threshold(-0.49, 1)
            drop_score = mm_2(-drop_score)
            drop_decision = drop_score.clone().requires_grad_()
            # print('rate of left edges', drop_decision.sum().data/drop_decision.shape[0])
            drop_matrix = lil_matrix((n_node, n_node), dtype=np.float32)
            drop_matrix[row, col] = drop_decision.cpu().data.numpy().squeeze(-1)
            att_dense_norm = att_dense_norm.multiply(drop_matrix.tocsr())  # update, remove the 0 edges

        if att_dense_norm[0, 0] == 0:  # add the weights of self-loop only add self-loop at the first layer
            degree = (att_dense_norm != 0).sum(1).A1
            lam = 1 / (degree + 1) # degree +1 is to add itself
            self_weight = sp.diags(np.array(lam), offsets=0, format="lil")
            att = att_dense_norm + self_weight  # add the self loop
        else:
            att = att_dense_norm

        row, col = att.nonzero()
        att_adj = np.vstack((row, col))
        att_edge_weight = att[row, col]
        att_edge_weight = np.exp(att_edge_weight)   # exponent, kind of softmax
        att_edge_weight = torch.tensor(np.array(att_edge_weight)[0], dtype=torch.float32)#.cuda()
        att_adj = torch.tensor(att_adj, dtype=torch.int64)#.cuda()

        shape = (n_node, n_node)
        new_adj = torch.sparse.FloatTensor(att_adj, att_edge_weight, shape)
        return new_adj

    def add_loop_sparse(self, adj, fill_value=1):
        # make identify sparse tensor
        row = torch.range(0, int(adj.shape[0]-1), dtype=torch.int64)
        i = torch.stack((row, row), dim=0)
        v = torch.ones(adj.shape[0], dtype=torch.float32)
        shape = adj.shape
        I_n = torch.sparse.FloatTensor(i, v, shape)
        return adj + I_n.to(self.device)


    def fit(self, features, adj, labels, idx_train, idx_val=None, idx_test=None, train_iters=81, att_0=None, attention=False, model_name=None, verbose=False, normalize=False, patience=510, ):
        '''
            train the gcn model, when idx_val is not None, pick the best model
            according to the validation loss
        '''
        sd = self.state_dict()
        for v in sd.values():
            self.device = v.device
            break
        self.sim = None
        self.idx_test = idx_test
        self.attention = attention
        # if self.attention:
        #     att_0 = self.att_coef_1(features, adj)
        #     adj = att_0 # update adj
        #     self.sim = att_0 # update att_0

        # self.device = self.gc1.weight.device

        if type(adj) is not torch.Tensor:
            features, adj, labels = utils.to_tensor(features, adj, labels, device=self.device)
        else:
            features = features.to(self.device)
            adj = adj.to(self.device)
            labels = labels.to(self.device)

        # normalize = False # we don't need normalize here, the norm is conducted in the GCN (self.gcn1) model
        # if normalize:
        #     if utils.is_sparse_tensor(adj):
        #         adj_norm = utils.normalize_adj_tensor(adj, sparse=True)
        #     else:
        #         adj_norm = utils.normalize_adj_tensor(adj)
        # else:
        #     adj_norm = adj
        # add self loop
        adj = self.add_loop_sparse(adj)


        """The normalization gonna be done in the GCNConv"""
        self.adj_norm = adj
        self.features = features
        self.labels = labels

        if idx_val is None:
            self._train_without_val(labels, idx_train, train_iters, verbose)
        else:
            if patience < train_iters:
                self._train_with_early_stopping(labels, idx_train, idx_val, train_iters, patience, verbose)
            else:
                self._train_with_val(labels, idx_train, idx_val, train_iters, verbose)

    def _train_without_val(self, labels, idx_train, train_iters, verbose):
        self.train()
        optimizer = optim.Adam(self.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        for i in range(train_iters):
            optimizer.zero_grad()
            output = self.forward(self.features, self.adj_norm)
            loss_train = F.nll_loss(output[idx_train], labels[idx_train], weight=None)   # this weight is the weight of each training nodes
            loss_train.backward()
            optimizer.step()
            if verbose and i % 20 == 0:
                print('Epoch {}, training loss: {}'.format(i, loss_train.item()))

        self.eval()
        output = self.forward(self.features, self.adj_norm)
        self.output = output

    def _train_with_val(self, labels, idx_train, idx_val, train_iters, verbose):
        if verbose:
            print('=== training gcn model ===')
        optimizer = optim.Adam(self.parameters(), lr=self.lr, weight_decay=self.weight_decay)

        best_loss_val = 100
        best_acc_val = 0

        for i in range(train_iters):
            # print('epoch', i)
            self.train()
            optimizer.zero_grad()
            output = self.forward(self.features, self.adj_norm)
            loss_train = F.nll_loss(output[idx_train], labels[idx_train])
            loss_train.backward()
            optimizer.step()
            self.eval()

            loss_val = F.nll_loss(output[idx_val], labels[idx_val])
            acc_val = utils.accuracy(output[idx_val], labels[idx_val])
            # acc_test = utils.accuracy(output[self.idx_test], labels[self.idx_test])

            # if verbose and i % 5 == 0:
            #     print('Epoch {}, training loss: {}, val acc: {}, '.format(i, loss_train.item(), acc_val))

            if best_loss_val > loss_val:
                best_loss_val = loss_val
                self.output = output
                weights = deepcopy(self.state_dict())

            if acc_val > best_acc_val:
                best_acc_val = acc_val
                self.output = output
                weights = deepcopy(self.state_dict())

        if verbose:
            print('=== picking the best model according to the performance on validation ===')
        self.load_state_dict(weights)
        # """my test"""
        # output_ = self.forward(self.features, self.adj_norm)
        # acc_test_ = utils.accuracy(output_[self.idx_test], labels[self.idx_test])
        # print('With best weights, test acc:', acc_test_)

    def _train_with_early_stopping(self, labels, idx_train, idx_val, train_iters, patience, verbose):
        if verbose:
            print('=== training gcn model ===')
        optimizer = optim.Adam(self.parameters(), lr=self.lr, weight_decay=self.weight_decay)

        early_stopping = patience
        best_loss_val = 100

        for i in range(train_iters):
            self.train()
            optimizer.zero_grad()
            output = self.forward(self.features, self.adj_norm)
            loss_train = F.nll_loss(output[idx_train], labels[idx_train])
            loss_train.backward()
            optimizer.step()

            self.eval()
            output = self.forward(self.features, self.adj_norm)

            if verbose and i % 10 == 0:
                print('Epoch {}, training loss: {}'.format(i, loss_train.item()))


            loss_val = F.nll_loss(output[idx_val], labels[idx_val])

            if best_loss_val > loss_val:
                best_loss_val = loss_val
                self.output = output
                weights = deepcopy(self.state_dict())
                patience = early_stopping
            else:
                patience -= 1
            if i > early_stopping and patience <= 0:
                break

        if verbose:
             print('=== early stopping at {0}, loss_val = {1} ==='.format(i, best_loss_val) )
        self.load_state_dict(weights)

    def test(self, idx_test):
        self.eval()
        output = self.predict()  # here use the self.features and self.adj_norm in training stage
        loss_test = F.nll_loss(output[idx_test], self.labels[idx_test])
        acc_test = utils.accuracy(output[idx_test], self.labels[idx_test])
        # print("Test set results:",
        #       "loss= {:.4f}".format(loss_test.item()),
        #       "accuracy= {:.4f}".format(acc_test.item()))
        return acc_test, output

    def _set_parameters(self):
        # TODO
        pass

    def predict(self, features=None, adj=None):
        '''By default, inputs are unnormalized data'''
        self.eval()
        if features is None and adj is None:
            return self.forward(self.features, self.adj_norm)
        else:
            if type(adj) is not torch.Tensor:
                features, adj = utils.to_tensor(features, adj, device=self.device)

            self.features = features
            if utils.is_sparse_tensor(adj):
                self.adj_norm = utils.normalize_adj_tensor(adj, sparse=True)
            else:
                self.adj_norm = utils.normalize_adj_tensor(adj)
            return self.forward(self.features, self.adj_norm)

class GCN4GNNGuard_attack(GCN):
    # Based on the existing GCN, add the robust part.
    def __init__(self, nfeat, nclass, nhid, activation, dropout=0.5, lr=0.01, drop=False, weight_decay=5e-4, with_relu=True, with_bias=True, add_self_loops = True, normalize = True):
        super(GCN4GNNGuard_attack, self).__init__(nfeat, nclass, nhid, activation, dropout=dropout, add_self_loops = add_self_loops, normalize = normalize)

        self.lr = lr
        self.weight_decay = weight_decay
        self.dropout = dropout
        self.nclass = nclass
        self.nfeat = nfeat
        self.hidden_sizes = nhid
        self.drop = drop
        if not with_relu:
            self.weight_decay = 0
        else:
            self.weight_decay = weight_decay
        self.with_relu = with_relu
        self.with_bias = with_bias

        self.gc1 = GCNConv(nfeat, nhid[0], bias=True,)
        self.gc2 = GCNConv(nhid[0], nclass, bias=True, )

    def forward(self, x, adj_lil):
        """we don't change the edge_index, just update the edge_weight;
        some edge_weight are regarded as removed if it equals to zero"""
        x = x.to_dense()
        adj = adj_lil.coalesce().indices()
        edge_weight = adj_lil.coalesce().values()

        x = F.relu(self.gc1(x, adj, edge_weight=edge_weight))
        x = F.dropout(x, self.dropout, training=self.training)
        x = self.gc2(x, adj, edge_weight=edge_weight)

        return F.log_softmax(x, dim=1)

    def add_loop_sparse(self, adj, fill_value=1):
        # make identify sparse tensor
        row = torch.range(0, int(adj.shape[0]-1), dtype=torch.int64)
        i = torch.stack((row, row), dim=0)
        v = torch.ones(adj.shape[0], dtype=torch.float32)
        shape = adj.shape
        I_n = torch.sparse.FloatTensor(i, v, shape)
        return adj + I_n.to(self.device)

    def initialize(self):
        self.gc1.reset_parameters()
        self.gc2.reset_parameters()

    def fit(self, features, adj, labels, idx_train, idx_val=None, idx_test=None, train_iters=81, att_0=None, attention=False, model_name=None, initialize=True, verbose=False, normalize=False, patience=510, ):
        '''
            train the gcn model, when idx_val is not None, pick the best model
            according to the validation loss
        '''
        sd = self.state_dict()
        for v in sd.values():
            self.device = v.device
            break

        self.sim = None
        self.attention = attention
        if self.attention:
            att_0 = self.att_coef_1(features, adj)
            adj = att_0 # update adj
            self.sim = att_0 # update att_0

        self.idx_test = idx_test
        
        if initialize:
            self.initialize()

        if type(adj) is not torch.Tensor:
            features, adj, labels = utils.to_tensor(features, adj, labels, device=self.device)
        else:
            features = features.to(self.device)
            adj = adj.to(self.device)
            labels = labels.to(self.device)

        normalize = False # we don't need normalize here, the norm is conducted in the GCN (self.gcn1) model
        if normalize:
            if utils.is_sparse_tensor(adj):
                adj_norm = utils.normalize_adj_tensor(adj, sparse=True)
            else:
                adj_norm = utils.normalize_adj_tensor(adj)
        else:
            adj_norm = adj
        # add self loop
        # adj = self.add_loop_sparse(adj)


        """Make the coefficient D^{-1/2}(A+I)D^{-1/2}"""
        self.adj_norm = adj_norm
        self.features = features
        self.labels = labels

        if idx_val is None:
            self._train_without_val(labels, idx_train, train_iters, verbose)
        else:
            if patience < train_iters:
                self._train_with_early_stopping(labels, idx_train, idx_val, train_iters, patience, verbose)
            else:
                self._train_with_val(labels, idx_train, idx_val, train_iters, verbose)

    def _train_without_val(self, labels, idx_train, train_iters, verbose):
        self.train()
        optimizer = optim.Adam(self.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        for i in range(train_iters):
            optimizer.zero_grad()
            output = self.forward(self.features, self.adj_norm)
            loss_train = F.nll_loss(output[idx_train], labels[idx_train], weight=None)   # this weight is the weight of each training nodes
            loss_train.backward()
            optimizer.step()
            if verbose and i % 10 == 0:
                print('Epoch {}, training loss: {}'.format(i, loss_train.item()))

        self.eval()
        output = self.forward(self.features, self.adj_norm)
        self.output = output

    def _train_with_val(self, labels, idx_train, idx_val, train_iters, verbose):
        if verbose:
            print('=== training gcn model ===')
        optimizer = optim.Adam(self.parameters(), lr=self.lr, weight_decay=self.weight_decay)

        best_loss_val = 100
        best_acc_val = 0

        for i in range(train_iters):
            # print('epoch', i)
            self.train()
            optimizer.zero_grad()
            output = self.forward(self.features, self.adj_norm)
            loss_train = F.nll_loss(output[idx_train], labels[idx_train])
            loss_train.backward()
            optimizer.step()

            acc_test =accuracy(output[self.idx_test], labels[self.idx_test])

            self.eval()
            output = self.forward(self.features, self.adj_norm)
            loss_val = F.nll_loss(output[idx_val], labels[idx_val])
            acc_val = utils.accuracy(output[idx_val], labels[idx_val])

            if verbose and i % 200 == 0:
                print('Epoch {}, training loss: {}, test acc: {}'.format(i, loss_train.item(), acc_test))

            if best_loss_val > loss_val:
                best_loss_val = loss_val
                self.output = output
                weights = deepcopy(self.state_dict())

            if acc_val > best_acc_val:
                best_acc_val = acc_val
                self.output = output
                weights = deepcopy(self.state_dict())

        if verbose:
            print('=== picking the best model according to the performance on validation ===')
        self.load_state_dict(weights)
        

    def _train_with_early_stopping(self, labels, idx_train, idx_val, train_iters, patience, verbose):
        if verbose:
            print('=== training gcn model ===')
        optimizer = optim.Adam(self.parameters(), lr=self.lr, weight_decay=self.weight_decay)

        early_stopping = patience
        best_loss_val = 100

        for i in range(train_iters):
            self.train()
            optimizer.zero_grad()
            output = self.forward(self.features, self.adj_norm)
            loss_train = F.nll_loss(output[idx_train], labels[idx_train])
            loss_train.backward()
            optimizer.step()

            self.eval()
            output = self.forward(self.features, self.adj_norm)

            if verbose and i % 10 == 0:
                print('Epoch {}, training loss: {}'.format(i, loss_train.item()))


            loss_val = F.nll_loss(output[idx_val], labels[idx_val])

            if best_loss_val > loss_val:
                best_loss_val = loss_val
                self.output = output
                weights = deepcopy(self.state_dict())
                patience = early_stopping
            else:
                patience -= 1
            if i > early_stopping and patience <= 0:
                break

        if verbose:
             print('=== early stopping at {0}, loss_val = {1} ==='.format(i, best_loss_val) )
        self.load_state_dict(weights)

    def test(self, idx_test):
        self.eval()
        output = self.predict()  # here use the self.features and self.adj_norm in training stage
        loss_test = F.nll_loss(output[idx_test], self.labels[idx_test])
        acc_test = utils.accuracy(output[idx_test], self.labels[idx_test])
        print("Test set results:",
              "loss= {:.4f}".format(loss_test.item()),
              "accuracy= {:.4f}".format(acc_test.item()))
        return acc_test, output

    def _set_parameters(self):
        # TODO
        pass

    def predict(self, features=None, adj=None):
        '''By default, inputs are unnormalized data'''
        # self.eval()
        if features is None and adj is None:
            return self.forward(self.features, self.adj_norm)
        else:
            if type(adj) is not torch.Tensor:
                features, adj = utils.to_tensor(features, adj, device=self.device)

            self.features = features
            if utils.is_sparse_tensor(adj):
                self.adj_norm = utils.normalize_adj_tensor(adj, sparse=True)
            else:
                self.adj_norm = utils.normalize_adj_tensor(adj)
            return self.forward(self.features, self.adj_norm)


@register_model("gnnguard-model")
class AutoGNNGuard(BaseAutoModel):
    def __init__(
        self,
        num_features: int = ...,
        num_classes: int = ...,
        device: _typing.Union[str, torch.device] = ...,
        **kwargs
    ) -> None:
        super().__init__(num_features, num_classes, device, **kwargs)
        self.hyper_parameter_space = [
            {
                "parameterName": "add_self_loops",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "normalize",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [128, 128, 128],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]

        self.hyper_parameters = {
            "num_layers": 3,
            "hidden": [128, 64],
            "dropout": 0,
            "act": "relu",
        }

    def _initialize(self):
        self._model = GCN4GNNGuard(
            nfeat = self.input_dimension,
            nclass = self.output_dimension,
            nhid = self.hyper_parameters.get("hidden"),
            activation = self.hyper_parameters.get("act"),
            dropout = self.hyper_parameters.get("dropout", None),
            add_self_loops = bool(self.hyper_parameters.get("add_self_loops", True)),
            normalize = bool(self.hyper_parameters.get("normalize", True)),
        ).to(self.device)

@register_model("gnnguard-attack-model")
class AutoGNNGuard_attack(BaseAutoModel):
    def __init__(
        self,
        num_features: int = ...,
        num_classes: int = ...,
        device: _typing.Union[str, torch.device] = ...,
        **kwargs
    ) -> None:
        super().__init__(num_features, num_classes, device, **kwargs)
        self.hyper_parameter_space = [
            {
                "parameterName": "add_self_loops",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "normalize",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [128, 128, 128],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]

        self.hyper_parameters = {
            "num_layers": 3,
            "hidden": [128, 64],
            "dropout": 0,
            "act": "relu",
        }

    def _initialize(self):
        self._model = GCN4GNNGuard_attack(
            nfeat = self.input_dimension,
            nclass = self.output_dimension,
            nhid = self.hyper_parameters.get("hidden"),
            activation = self.hyper_parameters.get("act"),
            dropout = self.hyper_parameters.get("dropout", None),
            add_self_loops = bool(self.hyper_parameters.get("add_self_loops", True)),
            normalize = bool(self.hyper_parameters.get("normalize", True)),
        ).to(self.device)


import os
import typing as _typing


class OnlineDataSource:
    @property
    def _raw_directory(self) -> str:
        return os.path.join(self.__path, "raw")

    @property
    def _processed_directory(self) -> str:
        return os.path.join(self.__path, "processed")

    @property
    def _raw_filenames(self) -> _typing.Iterable[str]:
        raise NotImplementedError

    @property
    def _processed_filenames(self) -> _typing.Iterable[str]:
        raise NotImplementedError

    @property
    def _raw_file_paths(self) -> _typing.Iterable[str]:
        return [
            os.path.join(self._raw_directory, raw_filename)
            for raw_filename in self._raw_filenames
        ]

    @property
    def _processed_file_paths(self) -> _typing.Iterable[str]:
        return [
            os.path.join(self._processed_directory, processed_filename)
            for processed_filename in self._processed_filenames
        ]

    @classmethod
    def __files_exist(cls, files: _typing.Iterable[str]) -> bool:
        return all([os.path.exists(file) for file in files])

    @classmethod
    def __make_directory(cls, path):
        os.makedirs(os.path.expanduser(os.path.normpath(path)), exist_ok=True)

    def _fetch(self):
        raise NotImplementedError

    def __fetch(self):
        if not self.__files_exist(self._raw_file_paths):
            self.__make_directory(self._raw_directory)
            self._fetch()

    def _process(self):
        raise NotImplementedError

    def __preprocess(self):
        if not self.__files_exist(self._processed_file_paths):
            self.__make_directory(self._processed_directory)
            self._process()

    def __getitem__(self, index: int) -> _typing.Any:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __init__(self, path: str):
        self.__path: str = os.path.expanduser(os.path.normpath(path))
        self.__fetch()
        self.__preprocess()



import typing as _typing
from autogl.utils import universal_registry
from ._base_feature_engineer import BaseFeatureEngineer


class _FeatureEngineerUniversalRegistryMetaclass(type):
    def __new__(
            mcs, name: str, bases: _typing.Tuple[type, ...],
            namespace: _typing.Dict[str, _typing.Any]
    ):
        return super(_FeatureEngineerUniversalRegistryMetaclass, mcs).__new__(
            mcs, name, bases, namespace
        )

    def __init__(
            cls, name: str, bases: _typing.Tuple[type, ...],
            namespace: _typing.Dict[str, _typing.Any]
    ):
        super(_FeatureEngineerUniversalRegistryMetaclass, cls).__init__(
            name, bases, namespace
        )
        cls._feature_engineer_universal_registry: _typing.MutableMapping[
            str, _typing.Type[BaseFeatureEngineer]
        ] = {}


class FeatureEngineerUniversalRegistry(universal_registry.UniversalRegistryBase):
    @classmethod
    def register_feature_engineer(cls, name: str) -> _typing.Callable[
        [_typing.Type[BaseFeatureEngineer]], _typing.Type[BaseFeatureEngineer]
    ]:
        def register_fe(fe: _typing.Type[BaseFeatureEngineer]) -> _typing.Type[BaseFeatureEngineer]:
            if not issubclass(fe, BaseFeatureEngineer):
                raise TypeError
            else:
                cls[name] = fe
                return fe

        return register_fe

    @classmethod
    def get_feature_engineer(cls, name: str) -> _typing.Type[BaseFeatureEngineer]:
        if name not in cls:
            raise ValueError(f"cannot find feature engineer {name}")
        else:
            return cls[name]


FEATURE_DICT = FeatureEngineerUniversalRegistry



from ..base import BaseFeature
import numpy as np


class BaseSelector(BaseFeature):
    def __init__(self, data_t="np", multigraph=False, **kwargs):
        super(BaseSelector, self).__init__(
            data_t=data_t, multigraph=multigraph, **kwargs
        )
        self._sel = None

    def _transform(self, data):
        if self._sel is not None:
            data.x = data.x[:, self._sel]
        return data



import typing as _typing
import torch.nn.functional
from torch_geometric.nn.conv import MessagePassing
from torch_sparse import SparseTensor, matmul

from . import register_model
from .base import ClassificationModel, ClassificationSupportedSequentialModel


class _GraphSAINTAggregationLayers:
    class MultiOrderAggregationLayer(torch.nn.Module):
        class Order0Aggregator(torch.nn.Module):
            def __init__(
                self,
                input_dimension: int,
                output_dimension: int,
                bias: bool = True,
                activation: _typing.Optional[str] = "ReLU",
                batch_norm: bool = True,
            ):
                super().__init__()
                if not type(input_dimension) == type(output_dimension) == int:
                    raise TypeError
                if not (input_dimension > 0 and output_dimension > 0):
                    raise ValueError
                if not type(bias) == bool:
                    raise TypeError
                self.__linear_transform = torch.nn.Linear(
                    input_dimension, output_dimension, bias
                )
                self.__linear_transform.reset_parameters()
                if type(activation) == str:
                    if activation.lower() == "ReLU".lower():
                        self.__activation = torch.nn.functional.relu
                    elif activation.lower() == "elu":
                        self.__activation = torch.nn.functional.elu
                    elif hasattr(torch.nn.functional, activation) and callable(
                        getattr(torch.nn.functional, activation)
                    ):
                        self.__activation = getattr(torch.nn.functional, activation)
                    else:
                        self.__activation = lambda x: x
                else:
                    self.__activation = lambda x: x
                if type(batch_norm) != bool:
                    raise TypeError
                else:
                    self.__optional_batch_normalization: _typing.Optional[
                        torch.nn.BatchNorm1d
                    ] = (
                        torch.nn.BatchNorm1d(output_dimension, 1e-8)
                        if batch_norm
                        else None
                    )

            def forward(
                self,
                x: _typing.Union[
                    torch.Tensor, _typing.Tuple[torch.Tensor, torch.Tensor]
                ],
                _edge_index: torch.Tensor,
                _edge_weight: _typing.Optional[torch.Tensor] = None,
                _size: _typing.Optional[_typing.Tuple[int, int]] = None,
            ) -> torch.Tensor:
                __output: torch.Tensor = self.__linear_transform(x)
                if self.__activation is not None and callable(self.__activation):
                    __output: torch.Tensor = self.__activation(__output)
                if self.__optional_batch_normalization is not None and isinstance(
                    self.__optional_batch_normalization, torch.nn.BatchNorm1d
                ):
                    __output: torch.Tensor = self.__optional_batch_normalization(
                        __output
                    )
                return __output

        class Order1Aggregator(MessagePassing):
            def __init__(
                self,
                input_dimension: int,
                output_dimension: int,
                bias: bool = True,
                activation: _typing.Optional[str] = "ReLU",
                batch_norm: bool = True,
            ):
                super().__init__(aggr="add")
                if not type(input_dimension) == type(output_dimension) == int:
                    raise TypeError
                if not (input_dimension > 0 and output_dimension > 0):
                    raise ValueError
                if not type(bias) == bool:
                    raise TypeError
                self.__linear_transform = torch.nn.Linear(
                    input_dimension, output_dimension, bias
                )
                self.__linear_transform.reset_parameters()
                if type(activation) == str:
                    if activation.lower() == "ReLU".lower():
                        self.__activation = torch.nn.functional.relu
                    elif activation.lower() == "elu":
                        self.__activation = torch.nn.functional.elu
                    elif hasattr(torch.nn.functional, activation) and callable(
                        getattr(torch.nn.functional, activation)
                    ):
                        self.__activation = getattr(torch.nn.functional, activation)
                    else:
                        self.__activation = lambda x: x
                else:
                    self.__activation = lambda x: x
                if type(batch_norm) != bool:
                    raise TypeError
                else:
                    self.__optional_batch_normalization: _typing.Optional[
                        torch.nn.BatchNorm1d
                    ] = (
                        torch.nn.BatchNorm1d(output_dimension, 1e-8)
                        if batch_norm
                        else None
                    )

            def forward(
                self,
                x: _typing.Union[
                    torch.Tensor, _typing.Tuple[torch.Tensor, torch.Tensor]
                ],
                _edge_index: torch.Tensor,
                _edge_weight: _typing.Optional[torch.Tensor] = None,
                _size: _typing.Optional[_typing.Tuple[int, int]] = None,
            ) -> torch.Tensor:

                if type(x) == torch.Tensor:
                    x: _typing.Tuple[torch.Tensor, torch.Tensor] = (x, x)

                __output = self.propagate(
                    _edge_index, x=x, edge_weight=_edge_weight, size=_size
                )
                __output: torch.Tensor = self.__linear_transform(__output)
                if self.__activation is not None and callable(self.__activation):
                    __output: torch.Tensor = self.__activation(__output)
                if self.__optional_batch_normalization is not None and isinstance(
                    self.__optional_batch_normalization, torch.nn.BatchNorm1d
                ):
                    __output: torch.Tensor = self.__optional_batch_normalization(
                        __output
                    )
                return __output

            def message(
                self, x_j: torch.Tensor, edge_weight: _typing.Optional[torch.Tensor]
            ) -> torch.Tensor:
                return x_j if edge_weight is None else edge_weight.view(-1, 1) * x_j

            def message_and_aggregate(
                self,
                adj_t: SparseTensor,
                x: _typing.Union[
                    torch.Tensor, _typing.Tuple[torch.Tensor, torch.Tensor]
                ],
            ) -> torch.Tensor:
                return matmul(adj_t, x[0], reduce=self.aggr)

        @property
        def integral_output_dimension(self) -> int:
            return (self._order + 1) * self._each_order_output_dimension

        def __init__(
            self,
            _input_dimension: int,
            _each_order_output_dimension: int,
            _order: int,
            bias: bool = True,
            activation: _typing.Optional[str] = "ReLU",
            batch_norm: bool = True,
            _dropout: _typing.Optional[float] = ...,
        ):
            super().__init__()
            if not (
                type(_input_dimension) == type(_order) == int
                and type(_each_order_output_dimension) == int
            ):
                raise TypeError
            if _input_dimension <= 0 or _each_order_output_dimension <= 0:
                raise ValueError
            if _order not in (0, 1):
                raise ValueError("Unsupported order number")
            self._input_dimension: int = _input_dimension
            self._each_order_output_dimension: int = _each_order_output_dimension
            self._order: int = _order
            if type(bias) != bool:
                raise TypeError
            self.__order0_transform = self.Order0Aggregator(
                self._input_dimension,
                self._each_order_output_dimension,
                bias,
                activation,
                batch_norm,
            )
            if _order == 1:
                self.__order1_transform = self.Order1Aggregator(
                    self._input_dimension,
                    self._each_order_output_dimension,
                    bias,
                    activation,
                    batch_norm,
                )
            else:
                self.__order1_transform = None
            if _dropout is not None and type(_dropout) == float:
                if _dropout < 0:
                    _dropout = 0
                if _dropout > 1:
                    _dropout = 1
                self.__optional_dropout: _typing.Optional[
                    torch.nn.Dropout
                ] = torch.nn.Dropout(_dropout)
            else:
                self.__optional_dropout: _typing.Optional[torch.nn.Dropout] = None

        def _forward(
            self,
            x: _typing.Union[torch.Tensor, _typing.Tuple[torch.Tensor, torch.Tensor]],
            edge_index: torch.Tensor,
            edge_weight: _typing.Optional[torch.Tensor] = None,
            size: _typing.Optional[_typing.Tuple[int, int]] = None,
        ) -> torch.Tensor:
            if self.__order1_transform is not None and isinstance(
                self.__order1_transform, self.Order1Aggregator
            ):
                __output: torch.Tensor = torch.cat(
                    [
                        self.__order0_transform(x, edge_index, edge_weight, size),
                        self.__order1_transform(x, edge_index, edge_weight, size),
                    ],
                    dim=1,
                )
            else:
                __output: torch.Tensor = self.__order0_transform(
                    x, edge_index, edge_weight, size
                )
            if self.__optional_dropout is not None and isinstance(
                self.__optional_dropout, torch.nn.Dropout
            ):
                __output: torch.Tensor = self.__optional_dropout(__output)
            return __output

        def forward(self, data) -> torch.Tensor:
            x: torch.Tensor = getattr(data, "x")
            if type(x) != torch.Tensor:
                raise TypeError
            edge_index: torch.LongTensor = getattr(data, "edge_index")
            if type(edge_index) != torch.Tensor:
                raise TypeError
            edge_weight: _typing.Optional[torch.Tensor] = getattr(
                data, "edge_weight", None
            )
            if edge_weight is not None and type(edge_weight) != torch.Tensor:
                raise TypeError
            return self._forward(x, edge_index, edge_weight)

    class WrappedDropout(torch.nn.Module):
        def __init__(self, dropout_module: torch.nn.Dropout):
            super().__init__()
            self.__dropout_module: torch.nn.Dropout = dropout_module

        def forward(self, tenser_or_data) -> torch.Tensor:
            if type(tenser_or_data) == torch.Tensor:
                return self.__dropout_module(tenser_or_data)
            elif (
                hasattr(tenser_or_data, "x")
                and type(getattr(tenser_or_data, "x")) == torch.Tensor
            ):
                return self.__dropout_module(getattr(tenser_or_data, "x"))
            else:
                raise TypeError


class GraphSAINTMultiOrderAggregationModel(ClassificationSupportedSequentialModel):
    def __init__(
        self,
        num_features: int,
        num_classes: int,
        _output_dimension_for_each_order: int,
        _layers_order_list: _typing.Sequence[int],
        _pre_dropout: float,
        _layers_dropout: _typing.Union[float, _typing.Sequence[float]],
        activation: _typing.Optional[str] = "ReLU",
        bias: bool = True,
        batch_norm: bool = True,
        normalize: bool = True,
    ):
        super(GraphSAINTMultiOrderAggregationModel, self).__init__()
        if type(_output_dimension_for_each_order) != int:
            raise TypeError
        if not _output_dimension_for_each_order > 0:
            raise ValueError
        self._layers_order_list: _typing.Sequence[int] = _layers_order_list

        if isinstance(_layers_dropout, _typing.Sequence):
            if len(_layers_dropout) != len(_layers_order_list):
                raise ValueError
            else:
                self._layers_dropout: _typing.Sequence[float] = _layers_dropout
        elif type(_layers_dropout) == float:
            if _layers_dropout < 0:
                _layers_dropout = 0
            if _layers_dropout > 1:
                _layers_dropout = 1
            self._layers_dropout: _typing.Sequence[float] = [
                _layers_dropout for _ in _layers_order_list
            ]
        else:
            raise TypeError
        if type(_pre_dropout) != float:
            raise TypeError
        else:
            if _pre_dropout < 0:
                _pre_dropout = 0
            if _pre_dropout > 1:
                _pre_dropout = 1
        self.__sequential_encoding_layers: torch.nn.ModuleList = torch.nn.ModuleList(
            (
                _GraphSAINTAggregationLayers.WrappedDropout(
                    torch.nn.Dropout(_pre_dropout)
                ),
                _GraphSAINTAggregationLayers.MultiOrderAggregationLayer(
                    num_features,
                    _output_dimension_for_each_order,
                    _layers_order_list[0],
                    bias,
                    activation,
                    batch_norm,
                    _layers_dropout[0],
                ),
            )
        )
        for _layer_index in range(1, len(_layers_order_list)):
            self.__sequential_encoding_layers.append(
                _GraphSAINTAggregationLayers.MultiOrderAggregationLayer(
                    self.__sequential_encoding_layers[-1].integral_output_dimension,
                    _output_dimension_for_each_order,
                    _layers_order_list[_layer_index],
                    bias,
                    activation,
                    batch_norm,
                    _layers_dropout[_layer_index],
                )
            )
        self.__apply_normalize: bool = normalize
        self.__linear_transform: torch.nn.Linear = torch.nn.Linear(
            self.__sequential_encoding_layers[-1].integral_output_dimension,
            num_classes,
            bias,
        )
        self.__linear_transform.reset_parameters()

    def cls_decode(self, x: torch.Tensor) -> torch.Tensor:
        if self.__apply_normalize:
            x: torch.Tensor = torch.nn.functional.normalize(x, p=2, dim=1)
        return torch.nn.functional.log_softmax(self.__linear_transform(x), dim=1)

    def cls_encode(self, data) -> torch.Tensor:
        if type(getattr(data, "x")) != torch.Tensor:
            raise TypeError
        if type(getattr(data, "edge_index")) != torch.Tensor:
            raise TypeError
        if (
            getattr(data, "edge_weight", None) is not None
            and type(getattr(data, "edge_weight")) != torch.Tensor
        ):
            raise TypeError
        for encoding_layer in self.__sequential_encoding_layers:
            setattr(data, "x", encoding_layer(data))
        return getattr(data, "x")

    @property
    def sequential_encoding_layers(self) -> torch.nn.ModuleList:
        return self.__sequential_encoding_layers


@register_model("GraphSAINTAggregationModel")
class GraphSAINTAggregationModel(ClassificationModel):
    def __init__(
        self,
        num_features: int = ...,
        num_classes: int = ...,
        device: _typing.Union[str, torch.device] = ...,
        init: bool = False,
        **kwargs
    ):
        super(GraphSAINTAggregationModel, self).__init__(
            num_features, num_classes, device=device, init=init, **kwargs
        )
        # todo: Initialize with default hyper parameter space and hyper parameter

    def _initialize(self):
        """ Initialize model """
        self.model = GraphSAINTMultiOrderAggregationModel(
            self.num_features,
            self.num_classes,
            self.hyper_parameter.get("output_dimension_for_each_order"),
            self.hyper_parameter.get("layers_order_list"),
            self.hyper_parameter.get("pre_dropout"),
            self.hyper_parameter.get("layers_dropout"),
            self.hyper_parameter.get("activation", "ReLU"),
            bool(self.hyper_parameter.get("bias", True)),
            bool(self.hyper_parameter.get("batch_norm", True)),
            bool(self.hyper_parameter.get("normalize", True)),
        ).to(self.device)



import numpy as np
import copy
from .base import BaseGenerator
from tqdm import tqdm
from ....utils import get_logger
from .. import register_feature

LOGGER = get_logger("Feature")


class Graphlet:
    def __init__(self, data, sample_error=0.1, sample_confidence=0.1):
        self._data = data
        self._init()

        self._sample_error = sample_error
        self._sample_confidence = sample_confidence
        self._dw = int(
            np.ceil(
                0.5 * (self._sample_error ** -2) * np.log(2 / self._sample_confidence)
            )
        )
        LOGGER.info(
            "sample error {} , confidence {},num {}".format(
                self._sample_error, self._sample_confidence, self._dw
            )
        )

    def _init(self):
        self._edges = list(self._data.edge_index)
        self._edges = [self._edges[0], self._edges[1]]
        self._num_nodes = self._data.x.shape[0]
        self._num_edges = len(self._edges[0])
        self._neighbours = [[] for _ in range(self._num_nodes)]
        for i in range(len(self._edges[0])):
            u, v = self._edges[0][i], self._edges[1][i]
            self._neighbours[u].append(v)

        LOGGER.info("nodes {} , edges {}".format(self._num_nodes, self._num_edges))

        # sorting
        self._node_degrees = np.array([len(x) for x in self._neighbours])
        self._nodes = np.argsort(self._node_degrees)
        for i in self._nodes:
            self._neighbours[i] = [
                x
                for _, x in sorted(
                    zip(self._node_degrees[self._neighbours[i]], self._neighbours[i]),
                    reverse=True,
                )
            ]
        self._neighbours = [np.array(x) for x in self._neighbours]

    def _get_gdv(self, v, u):
        if self._node_degrees[v] >= self._node_degrees[u]:
            pass
        else:
            u, v = v, u
        Sv, Su, Te = set(), set(), set()
        sigma1, sigma2 = 0, 0
        nb = self._neighbours
        N = self._num_nodes
        M = self._num_edges
        phi = np.zeros(self._num_nodes, dtype=int)
        c1, c2, c3, c4 = 1, 2, 3, 4
        x = np.zeros(16, dtype=int)
        # p1
        for w in nb[v]:
            if w != u:
                Sv.add(w)
                phi[w] = c1
        # p2
        for w in nb[u]:
            if w != v:
                if phi[w] == c1:
                    Te.add(w)
                    phi[w] = c3
                    Sv.remove(w)
                else:
                    Su.add(w)
                    phi[w] = c2
        # p3
        for w in Te:
            for r in nb[w]:
                if phi[r] == c3:
                    x[5] += 1
            phi[w] = c4
            sigma2 = sigma2 + len(nb[w]) - 2
        # p4
        for w in Su:
            for r in nb[w]:
                if phi[r] == c1:
                    x[8] += 1
                if phi[r] == c2:
                    x[7] += 1
                if phi[r] == c4:
                    sigma1 += 1
            phi[w] = 0
            sigma2 = sigma2 + len(nb[w]) - 1
        # p5
        for w in Sv:
            for r in nb[w]:
                if phi[r] == c1:
                    x[7] += 1
                if phi[r] == c4:
                    sigma1 += 1
            phi[w] = 0
            sigma2 = sigma2 + len(nb[w]) - 1

        lsv, lsu, lte, du, dv = len(Sv), len(Su), len(Te), len(nb[u]), len(nb[v])
        # 3-graphlet
        x[1] = lte
        x[2] = du + dv - 2 - 2 * x[1]
        x[3] = N - x[2] - x[1] - 2
        x[4] = N * (N - 1) * (N - 2) / 6 - (x[1] + x[2] + x[3])
        # 4 connected graphlets
        x[6] = x[1] * (x[1] - 1) / 2 - x[5]
        x[10] = lsv * lsu - x[8]
        x[9] = lsv * (lsv - 1) / 2 + lsu * (lsu - 1) / 2 - x[7]
        # 4 diconnected graphlets
        t1 = N - (lte + lsu + lsv + 2)
        x[11] = x[1] * t1
        x[12] = M - (du + dv - 1) - (sigma2 - sigma1 - x[5] - x[8] - x[7])
        x[13] = (lsu + lsv) * t1
        x[14] = t1 * (t1 - 1) / 2 - x[12]
        x[15] = N * (N - 1) * (N - 2) * (N - 3) / 24 - np.sum(x[5:15])

        return x

    def _get_gdv_sample(self, v, u):
        if self._node_degrees[v] >= self._node_degrees[u]:
            pass
        else:
            u, v = v, u
        Sv = set()
        sigma1, sigma2 = 0, 0
        nb = self._neighbours
        N = self._num_nodes
        M = self._num_edges
        phi = np.zeros(self._num_nodes, dtype=int)
        c1, c2, c3, c4 = 1, 2, 3, 4
        x = np.zeros(16)
        dw = self._dw

        # p1
        Sv = set(nb[v][nb[v] != u])
        phi[list(Sv)] = c1
        # p2
        p2w = nb[u][nb[u] != c1]
        p2w1 = p2w[phi[p2w] == c1]
        p2w2 = p2w[phi[p2w] != c1]
        Te = p2w1
        phi[p2w1] = c3
        Sv -= set(list(p2w1))
        Su = p2w2
        phi[p2w2] = c2
        # p3
        for w in Te:
            if dw >= len(nb[w]):
                region = nb[w]
                inc = 1
            else:
                region = np.random.choice(nb[w], dw, replace=False)
                inc = self._node_degrees[w] / dw
            phir = phi[region]
            x[5] += inc * np.sum(phir == c3)
            phi[w] = c4
            sigma2 = sigma2 + len(nb[w]) - 2
        # p4
        for w in Su:
            if dw >= len(nb[w]):
                region = nb[w]
                inc = 1
            else:
                region = np.random.choice(nb[w], dw, replace=False)
                inc = self._node_degrees[w] / dw
            phir = phi[region]
            x[8] += inc * np.sum(phir == c1)
            x[7] += inc * np.sum(phir == c2)
            sigma1 += inc * np.sum(phir == c4)
            phi[w] = 0
            sigma2 = sigma2 + len(nb[w]) - 1
        # p5
        for w in Sv:
            if dw >= len(nb[w]):
                region = nb[w]
                inc = 1
            else:
                region = np.random.choice(nb[w], dw, replace=False)
                inc = self._node_degrees[w] / dw
            phir = phi[region]
            x[7] += inc * np.sum(phir == c1)
            sigma1 += inc * np.sum(phir == c4)
            phi[w] = 0
            sigma2 = sigma2 + len(nb[w]) - 1

        lsv, lsu, lte, du, dv = len(Sv), len(Su), len(Te), len(nb[u]), len(nb[v])
        # 3-graphlet
        x[1] = lte
        x[2] = du + dv - 2 - 2 * x[1]
        x[3] = N - x[2] - x[1] - 2
        x[4] = N * (N - 1) * (N - 2) / 6 - (x[1] + x[2] + x[3])
        # 4 connected graphlets
        x[6] = x[1] * (x[1] - 1) / 2 - x[5]
        x[10] = lsv * lsu - x[8]
        x[9] = lsv * (lsv - 1) / 2 + lsu * (lsu - 1) / 2 - x[7]
        # 4 diconnected graphlets
        t1 = N - (lte + lsu + lsv + 2)
        x[11] = x[1] * t1
        x[12] = M - (du + dv - 1) - (sigma2 - sigma1 - x[5] - x[8] - x[7])
        x[13] = (lsu + lsv) * t1
        x[14] = t1 * (t1 - 1) / 2 - x[12]
        x[15] = N * (N - 1) * (N - 2) * (N - 3) / 24 - np.sum(x[5:15])

        return x

    def get_gdvs(self, sample=True):
        res = np.zeros((self._num_nodes, 15))
        for u in tqdm(range(self._num_nodes)):
            vs = self._neighbours[u]
            if len(vs) != 0:
                gdvs = []
                for v in tqdm(vs, disable=len(vs) < 100):
                    if sample:
                        gdvs.append(self._get_gdv_sample(u, v))
                    else:
                        gdvs.append(self._get_gdv(u, v))
                res[u, :] = np.mean(gdvs, axis=0)[1:]
        return res

    def get_gdvs_cp(self, workers="max"):
        r"""
        c++ parallel , same function as get_gdvs
        """
        tmpfile = "tmp.mtx"
        tmpmicro = "tmp.micro"
        self._save(tmpfile)
        os.system(
            "{} -f {} --micro {} -w {}".format(pgd_path, tmpfile, tmpmicro, workers)
        )
        return self._load(tmpmicro)

    def _save(self, filename):
        with open(filename, "w") as file:
            file.write(
                "{} {} {}\n".format(self._num_nodes, self._num_nodes, self._num_edges)
            )
            for u in self._nodes:
                for v in self._neighbours[u]:
                    file.write("{} {}\n".format(u + 1, v + 1))

    def _load(self, filename):
        df = pd.read_csv(filename)
        edges = df[["% src", "dst"]].values
        egdvs = df.values[:, 2:]

        num_nodes = np.max(edges)
        ngdvs = np.zeros((num_nodes, 8))

        nbs = [[] for _ in range(num_nodes)]
        for i, (u, v) in enumerate(edges):
            u -= 1
            v -= 1
            nbs[u].append(i)
            nbs[v].append(i)

        for i in range(num_nodes):
            if len(nbs[i]) != 0:
                ngdvs[i, :] = np.mean(egdvs[nbs[i]], axis=0)
        return ngdvs


@register_feature("graphlet")
class GeGraphlet(BaseGenerator):
    r"""generate local graphlet numbers as features. The implementation refers to [#]_ .

    References
    ----------
    .. [#] Ahmed, N. K., Willke, T. L., & Rossi, R. A. (2016).
        Estimation of local subgraph counts. Proceedings - 2016 IEEE International Conference on Big Data, Big Data 2016, 586–595.
        https://doi.org/10.1109/BigData.2016.7840651

    """

    def __init__(self, workers=1):
        super(GeGraphlet, self).__init__()
        self.workers = workers

    def _transform(self, data):
        r"""graphlet degree vectors"""
        gl = Graphlet(data)
        #         res=gl.get_gdvs_cp(self.workers)
        res = gl.get_gdvs()
        data.x = np.concatenate([data.x, res], axis=1)
        return data
